import math

from .AnimationClip import AABB, PackedFloatVector, PackedIntVector
from .NamedObject import NamedObject
from .Texture2D import StreamingInfo
from ..helpers.ResourceReader import get_resource_data
from ..math import Matrix4x4, Vector3
from ..streams import EndianBinaryWriter
from ..enums import GfxPrimitiveType
import struct
from enum import IntEnum
from ..export import MeshExporter


class MinMaxAABB:
    def __init__(self, reader):
        self.m_Min = reader.read_vector3()
        self.m_Max = reader.read_vector3()

    def save(self, writer):
        writer.write_vector3(self.m_Min)
        writer.write_vector3(self.m_Max)


class CompressedMesh:
    def __init__(self, reader):
        version = reader.version
        self.m_Vertices = PackedFloatVector(reader)
        self.m_UV = PackedFloatVector(reader)
        if version[0] < 5:  # 5 down
            self.m_BindPoses = PackedFloatVector(reader)
        self.m_Normals = PackedFloatVector(reader)
        self.m_Tangents = PackedFloatVector(reader)
        self.m_Weights = PackedIntVector(reader)
        self.m_NormalSigns = PackedIntVector(reader)
        self.m_TangentSigns = PackedIntVector(reader)
        if version >= (5,):  # 5 and up
            self.m_FloatColors = PackedFloatVector(reader)
        self.m_BoneIndices = PackedIntVector(reader)
        self.m_Triangles = PackedIntVector(reader)
        if version >= (3, 5):  # 3.5 and up
            if version[0] < 5:  # 5 down
                self.m_Colors = PackedIntVector(reader)
            else:
                self.m_UVInfo = reader.read_u_int()

    def save(self, writer, version):
        self.m_Vertices.save(writer)
        self.m_UV.save(writer)
        if version < (5,):  # 5 down
            self.m_BindPoses.save(writer)
        self.m_Normals.save(writer)
        self.m_Tangents.save(writer)
        self.m_Weights.save(writer)
        self.m_NormalSigns.save(writer)
        self.m_TangentSigns.save(writer)
        if version >= (5,):  # 5 and up
            self.m_FloatColors.save(writer)
        self.m_BoneIndices.save(writer)
        self.m_Triangles.save(writer)
        if version >= (3, 5):  # 3.5 and up
            if version < (5,):  # 5 down
                self.m_Colors.save(writer)
            else:
                writer.write_u_int(self.m_UVInfo)


class StreamInfo:
    def __init__(self, **kwargs):
        if "reader" in kwargs:
            reader = kwargs["reader"]
            version = reader.version
            self.channelMask = reader.read_u_int()
            self.offset = reader.read_u_int()

            if version < (4,):  # 4.0 down
                self.stride = reader.read_u_int()
                self.align = reader.read_u_int()
            else:
                self.stride = reader.read_byte()
                self.dividerOp = reader.read_byte()
                self.frequency = reader.read_u_short()
        else:
            self.__dict__ = kwargs

    def save(self, writer: EndianBinaryWriter, version: tuple):
        writer.write_u_int(self.channelMask)
        writer.write_u_int(self.offset)

        if version < (4,):  # 4.0 down
            writer.write_u_int(self.stride)
            writer.write_u_int(self.align)
        else:
            writer.write_byte(self.stride)
            writer.write_byte(self.dividerOp)
            writer.write_u_short(self.frequency)


class ChannelInfo:
    def __init__(self, reader):
        self.stream = reader.read_byte()
        self.offset = reader.read_byte()
        self.format = reader.read_byte()
        self.dimension = reader.read_byte()

    def save(self, writer):
        writer.write_byte(self.stream)
        writer.write_byte(self.offset)
        writer.write_byte(self.format)
        writer.write_byte(self.dimension)


class VertexData:
    def __init__(self, reader):
        self.reader = reader
        version = reader.version

        if version < (2018,):  # 2018 down
            self.m_CurrentChannels = reader.read_u_int()

        self.m_VertexCount = reader.read_u_int()

        if version >= (4,):  # 4.0 and up
            m_ChannelsSize = reader.read_int()
            self.m_Channels = [ChannelInfo(reader) for _ in range(m_ChannelsSize)]

        if version < (5,):  # 5.0 down
            if version < (4,):  # 4.0 down
                m_StreamsSize = 4
            else:
                m_StreamsSize = reader.read_int()

            self.m_Streams = [StreamInfo(reader=reader) for _ in range(m_StreamsSize)]

            if version < (4,):  # 4.0 down
                self.GetChannels()
        else:  # 5.0 and up
            self.GetStreams()

        self.m_DataSize = reader.read_bytes(reader.read_int())
        reader.align_stream()

    def save(self, writer: EndianBinaryWriter, version):
        if version < (2018,):  # 2018 down
            writer.write_u_int(self.m_CurrentChannels)

        writer.write_u_int(self.m_VertexCount)

        if version >= (4,):  # 4.0 and up
            writer.write_int(len(self.m_Channels))
            for ch in self.m_Channels:
                ch.save(writer)

        if (4,) <= version[:2] < (5,):  # 4.0 and up to 5.0
            writer.write_int(len(self.m_Streams))

            for stream in self.m_Streams:
                stream.save(writer=writer, version=version)

            if version < (4,):  # 4.0 down
                raise Exception("Unsupported version")
        else:  # 5.0 and up
            # for stream in self.m_Streams:
            #    stream.save(writer)
            pass

        writer.write_int(len(self.m_DataSize))
        writer.write_bytes(self.m_DataSize)
        writer.align_stream()

    def GetStreams(self):
        streamCount = 1
        if self.m_Channels:
            streamCount += max(x.stream for x in self.m_Channels)

        self.m_Streams = {}
        offset = 0
        for s in range(streamCount):
            chnMask = 0
            stride = 0
            for chn, m_Channel in enumerate(self.m_Channels):
                if m_Channel.stream == s:
                    if m_Channel.dimension > 0:
                        chnMask |= 1 << chn  # Shift 1UInt << chn
                        stride += m_Channel.dimension * MeshHelper.GetFormatSize(
                            self.reader.version, m_Channel.format
                        )
            self.m_Streams[s] = StreamInfo(
                channelMask=chnMask,
                offset=offset,
                stride=stride,
                dividerOp=0,
                frequency=0,
            )
            offset += self.m_VertexCount * stride
            # static size_t align_streamSize (size_t size) { return (size + (kVertexStreamAlign-1)) & ~(kVertexStreamAlign-1)
            offset = (offset + (16 - 1)) & ~(
                16 - 1
            )  # (offset + (16u - 1u)) & ~(16u - 1u);

    def GetChannels(self):
        self.m_Channels = []  # ChannelInfo[6]
        for i in range(6):
            self.m_Channels.append(ChannelInfo(self.reader))
        for s, m_Stream in enumerate(self.m_Streams):
            channelMask = bytearray(m_Stream.channelMask)  # BitArray
            offset = 0
            for i in range(6):
                if channelMask[i]:
                    m_Channel = self.m_Channels[i]
                    m_Channel.stream = s
                    m_Channel.offset = offset
                    if i in [0, 1]:
                        # 0 - kShaderChannelVertex
                        # 1 - kShaderChannelNormal
                        m_Channel.format = 0  # kChannelFormatFloat
                        m_Channel.dimension = 3
                    elif i == 2:  # kShaderChannelColor
                        m_Channel.format = 2  # kChannelFormatColor
                        m_Channel.dimension = 4
                    elif i in [3, 4]:
                        # 3 - kShaderChannelTexCoord0
                        # 4 - kShaderChannelTexCoord1
                        m_Channel.format = 0  # kChannelFormatFloat
                        m_Channel.dimension = 2
                    elif i == 5:  # kShaderChannelTangent
                        m_Channel.format = 0  # kChannelFormatFloat
                        m_Channel.dimension = 4
                    offset += m_Channel.dimension * MeshHelper.GetFormatSize(
                        self.reader.version, m_Channel.format
                    )


class BoneWeights4:
    def __init__(self, reader=None):
        if reader:
            self.weight = reader.read_float_array(4)
            self.boneIndex = reader.read_int_array(4)
        else:
            self.weight = [0.0] * 4
            self.boneIndex = [0] * 4

    def save(self, writer):
        writer.write_float_array(self.weight)
        writer.write_int_array(self.boneIndex)


class BlendShapeVertex:
    def __init__(self, reader):
        self.vertex = reader.read_vector3()
        self.normal = reader.read_vector3()
        self.tangent = reader.read_vector3()
        self.index = reader.read_u_int()


class MeshBlendShape:
    def __init__(self, reader):
        version = reader.version

        if version < (4, 3):  # 4.3 down
            self.name = reader.read_aligned_string()
        self.firstVertex = reader.read_u_int()
        self.vertexCount = reader.read_u_int()
        if version < (4, 3):  # 4.3 down
            self.aabbMinDelta = reader.read_vector3()
            self.aabbMaxDelta = reader.read_vector3()
        self.hasNormals = reader.read_boolean()
        self.hasTangents = reader.read_boolean()
        if version >= (4, 3):  # 4.3 and up
            reader.align_stream()


class MeshBlendShapeChannel:
    def __init__(self, reader):
        self.name = reader.read_aligned_string()
        self.nameHash = reader.read_u_int()
        self.frameIndex = reader.read_int()
        self.frameCount = reader.read_int()


class BlendShapeData:
    def __init__(self, reader):
        version = reader.version

        if version >= (4, 3):  # 4.3 and up
            numVerts = reader.read_int()
            self.vertices = [BlendShapeVertex(reader) for _ in range(numVerts)]

            numShapes = reader.read_int()
            self.shapes = [MeshBlendShape(reader) for _ in range(numShapes)]

            numChannels = reader.read_int()
            self.channels = [MeshBlendShapeChannel(reader) for _ in range(numChannels)]
            self.fullWeights = reader.read_float_array()
        else:
            m_ShapesSize = reader.read_int()
            self.m_Shapes = [MeshBlendShape(reader) for _ in range(m_ShapesSize)]
            reader.align_stream()
            m_ShapeVerticesSize = reader.read_int()
            self.m_ShapeVertices = [
                BlendShapeVertex(reader) for _ in range(m_ShapeVerticesSize)
            ]


class SubMesh:
    def __init__(self, reader):
        version = reader.version
        self.firstByte = reader.read_u_int()
        self.indexCount = reader.read_u_int()
        self.topology = GfxPrimitiveType(reader.read_int())

        if version < (4,):  # 4.0 down
            self.triangleCount = reader.read_u_int()

        if version >= (2017, 3):  # 2017.3 and up
            self.baseVertex = reader.read_u_int()

        if version >= (3,):  # 3.0 and up
            self.firstVertex = reader.read_u_int()
            self.vertexCount = reader.read_u_int()
            self.localAABB = AABB(reader)

    def save(self, writer, version):
        writer.write_u_int(self.firstByte)
        writer.write_u_int(self.indexCount)
        writer.write_int(self.topology.value)

        if version < (4,):  # 4.0 down
            writer.write_u_int(self.triangleCount)

        if version >= (2017, 3):  # 2017.3 and up
            writer.write_u_int(self.baseVertex)

        if version >= (3,):  # 3.0 and up
            writer.write_u_int(self.firstVertex)
            writer.write_u_int(self.vertexCount)
            self.localAABB.save(writer)


class Mesh(NamedObject):
    def export(self):
        return MeshExporter.export_mesh(self)

    def __init__(self, reader):
        super().__init__(reader=reader)
        version = reader.version

        self.m_Use16BitIndices = True
        self.m_Indices = []
        self.m_BindPose = []
        self.m_BoneNameHashes = []
        self.m_Vertices = []
        self.m_Skin = []
        self.m_Normals = []
        self.m_Colors = []
        self.m_UV0 = []
        self.m_UV1 = []
        self.m_UV2 = []
        self.m_UV3 = []
        self.m_UV4 = []
        self.m_UV5 = []
        self.m_UV6 = []
        self.m_UV7 = []
        self.m_Tangents = []

        if version < (3, 5):  # 3.5 down
            self.m_Use16BitIndices = reader.read_int() > 0

        if version[:2] <= (2, 5):  # 2.5 and down
            m_IndexBuffer_size = reader.read_int()

            if self.m_Use16BitIndices:
                self.m_IndexBuffer = [
                    reader.read_u_short()
                    for _ in range(math.ceil(m_IndexBuffer_size / 2))
                ]
                reader.align_stream()
            else:
                self.m_IndexBuffer = reader.read_u_int_array(
                    math.ceil(m_IndexBuffer_size / 4)
                )

        m_SubMeshesSize = reader.read_int()
        self.m_SubMeshes = [SubMesh(reader) for _ in range(m_SubMeshesSize)]

        if version >= (4, 1):  # 4.1 and up
            self.m_Shapes = BlendShapeData(reader)

        if version >= (4, 3):  # 4.3 and up
            self.m_BindPose = reader.read_matrix_array()
            self.m_BoneNameHashes = reader.read_u_int_array()
            self.m_RootBoneNameHash = reader.read_u_int()

        if version >= (2, 6):  # 2.6.0 and up
            if version >= (2019,):  # 2019 and up
                m_BonesAABBSize = reader.read_int()
                self.m_BonesAABB = [MinMaxAABB(reader) for _ in range(m_BonesAABBSize)]
                self.m_VariableBoneCountWeights = reader.read_u_int_array()

            self.m_MeshCompression = reader.read_byte()
            if version >= (4,):  #
                if version < (5,):  #
                    self.m_StreamCompression = reader.read_byte()
                self.m_IsReadable = reader.read_boolean()
                self.m_KeepVertices = reader.read_boolean()
                self.m_KeepIndices = reader.read_boolean()
            reader.align_stream()

            # Unity fixed it in 2017.3.1p1 and later versions
            if (
                version >= (2017, 4)  # 2017.4
                # fixed after 2017.3.1px
                or version[:3] == (2017, 3, 1)
                and self.build_type.IsPatch
                # 2017.3.xfx with no compression
                or version[:2] == (2017, 3)
                and self.m_MeshCompression == 0
            ):
                self.m_IndexFormat = reader.read_int()
                self.m_Use16BitIndices = self.m_IndexFormat == 0

            m_IndexBuffer_size = reader.read_int()
            if self.m_Use16BitIndices:
                self.m_IndexBuffer = [
                    reader.read_u_short()
                    for _ in range(math.ceil(m_IndexBuffer_size / 2))
                ]
                reader.align_stream()
            else:
                self.m_IndexBuffer = reader.read_u_int_array(
                    math.ceil(m_IndexBuffer_size / 4)
                )

        if version < (3, 5):  # 3.4.2 and earlier
            self.m_VertexCount = reader.read_int()
            self.m_Vertices = reader.read_float_array(self.m_VertexCount * 3)  # Vector3

            self.m_SkinSize = reader.read_int()
            self.m_Skin = [BoneWeights4(reader) for _ in range(self.m_SkinSize)]

            self.m_BindPose = reader.read_matrix_array()
            self.m_UV0 = reader.read_float_array(reader.read_int() * 2)  # Vector2
            self.m_UV1 = reader.read_float_array(reader.read_int() * 2)  # Vector2

            if version[:2] <= (2, 5):  # 2.5 and down
                m_TangentSpace_size = reader.read_int()
                self.m_Normals = [0] * (m_TangentSpace_size * 3)
                self.m_Tangets = [0] * (m_TangentSpace_size * 4)
                for v in range(m_TangentSpace_size):
                    self.m_Normals[v * 3] = reader.read_float()
                    self.m_Normals[v * 3 + 1] = reader.read_float()
                    self.m_Normals[v * 3 + 2] = reader.read_float()
                    self.m_Tangents[v * 3] = reader.read_float()
                    self.m_Tangents[v * 3 + 1] = reader.read_float()
                    self.m_Tangents[v * 3 + 2] = reader.read_float()
                    # handedness
                    self.m_Tangents[v * 3 + 3] = reader.read_float()
            else:  # 2.6.0 and later
                self.m_Tangents = reader.read_float_array(
                    reader.read_int() * 4
                )  # Vector4
                self.m_Normals = reader.read_float_array(
                    reader.read_int() * 3
                )  # Vector3
        else:
            if version[:2] < (2018, 2):  # 2018.2 down
                m_SkinSize = reader.read_int()
                self.m_Skin = [BoneWeights4(reader) for _ in range(m_SkinSize)]

            if version[:2] <= (4, 2):  # 4.2 and down
                self.m_BindPose = reader.read_matrix_array()

            self.m_VertexData = VertexData(reader)

        if version >= (2, 6):  # 2.6.0 and later
            self.m_CompressedMesh = CompressedMesh(reader)

        self.m_LocalAABB = AABB(reader)

        if version[:2] <= (3, 4):  # 3.4.2 and earlier
            m_Colors_size = reader.read_int()
            self.mColors = [reader.read_byte() / 0xFF for _ in range(m_Colors_size * 4)]

            m_CollisionTriangles_size = reader.read_int()
            reader.Position += m_CollisionTriangles_size * 4  # UInt32 indices
            m_CollisionVertexCount = reader.read_int()

        self.m_MeshUsageFlags = reader.read_int()
        if version >= (5,):  # 5.0 and up
            self.m_BakedConvexCollisionMesh = reader.read_bytes(reader.read_int())
            reader.align_stream()
            self.m_BakedTriangleCollisionMesh = reader.read_bytes(reader.read_int())
            reader.align_stream()

        if version >= (2018, 2):  # 2018.2 and up
            self.m_MeshMetrics = [reader.read_float(), reader.read_float()]

        if version >= (2018, 3):  # 2018.3 and up
            reader.align_stream()
            self.m_StreamData = StreamingInfo(reader, version)

            if self.m_StreamData.path and self.m_VertexData.m_VertexCount > 0:
                self.m_VertexData.m_DataSize = get_resource_data(
                    self.m_StreamData.path,
                    self.assets_file,
                    self.m_StreamData.offset,
                    self.m_StreamData.size,
                )

        # Fix channel after 2018.3
        version = self.version
        if version >= (3, 5):  # 3.5 and up
            self.ReadVertexData()

        if version >= (2, 6):  # 2.6.0 and later
            self.DecompressCompressedMesh()

        self.GetTriangles()

    def ReadVertexData(self):
        version = self.version
        m_VertexData = self.m_VertexData
        m_VertexCount = self.m_VertexCount = m_VertexData.m_VertexCount

        for chn, m_Channel in enumerate(m_VertexData.m_Channels):
            if m_Channel.dimension > 0:
                m_Stream = m_VertexData.m_Streams[m_Channel.stream]
                channelMask = bin(m_Stream.channelMask)[::-1]
                if channelMask[chn] == "1":
                    if version[0] < 2018 and chn == 2 and m_Channel.format == 2:
                        m_Channel.dimension = 4

                    componentByteSize = MeshHelper.GetFormatSize(
                        version, m_Channel.format
                    )
                    componentBytes = bytearray(
                        m_VertexCount * m_Channel.dimension * componentByteSize
                    )
                    for v in range(m_VertexCount):
                        vertexOffset = (
                            int(m_Stream.offset)
                            + m_Channel.offset
                            + int(m_Stream.stride) * v
                        )
                        for d in range(m_Channel.dimension):
                            componentOffsetSrc = (
                                vertexOffset + componentByteSize * d
                            )  # src offset
                            componentOffsetDst = componentByteSize * (
                                v * m_Channel.dimension + d
                            )  # dst offst

                            buff = m_VertexData.m_DataSize[
                                componentOffsetSrc : componentOffsetSrc
                                + componentByteSize
                            ]

                            if (
                                self.reader.endian == "<" and componentByteSize > 1
                            ):  # swap bytes
                                buff = buff[::-1]

                            componentBytes[
                                componentOffsetDst : componentOffsetDst
                                + componentByteSize
                            ] = buff
                            # Buffer.BlockCopy(m_VertexData.m_DataSize, componentOffset, componentBytes, componentByteSize * (v * m_Channel.dimension + d), componentByteSize);
                            # (Array src, int srcOffset, Array dst, int dstOffset, int count);

                    if MeshHelper.IsIntFormat(version, m_Channel.format):
                        componentsIntArray = MeshHelper.BytesToIntArray(
                            componentBytes, componentByteSize
                        )
                    else:
                        componentsFloatArray = MeshHelper.BytesToFloatArray(
                            componentBytes, componentByteSize
                        )

                    if version[0] >= 2018:
                        if chn == 0:  # kShaderChannelVertex
                            self.m_Vertices = componentsFloatArray
                        elif chn == 1:  # kShaderChannelNormal
                            self.m_Normals = componentsFloatArray
                        elif chn == 2:  # kShaderChannelTangent
                            self.m_Tangents = componentsFloatArray
                        elif chn == 3:  # kShaderChannelColor
                            self.m_Colors = componentsFloatArray
                        elif chn == 4:  # kShaderChannelTexCoord0
                            self.m_UV0 = componentsFloatArray
                        elif chn == 5:  # kShaderChannelTexCoord1
                            self.m_UV1 = componentsFloatArray
                        elif chn == 6:  # kShaderChannelTexCoord2
                            self.m_UV2 = componentsFloatArray
                        elif chn == 7:  # kShaderChannelTexCoord3
                            self.m_UV3 = componentsFloatArray
                        elif chn == 8:  # kShaderChannelTexCoord4
                            self.m_UV4 = componentsFloatArray
                        elif chn == 9:  # kShaderChannelTexCoord5
                            self.m_UV5 = componentsFloatArray
                        elif chn == 10:  # kShaderChannelTexCoord6
                            self.m_UV6 = componentsFloatArray
                        elif chn == 11:  # kShaderChannelTexCoord7
                            self.m_UV7 = componentsFloatArray
                        # 2018.2 and up
                        elif chn == 12:  # kShaderChannelBlendWeight
                            if not self.m_Skin:
                                self.InitMSkin()
                            for i in range(m_VertexCount):
                                for j in range(m_Channel.dimension):
                                    self.m_Skin[i].weight[j] = componentsFloatArray[
                                        i * m_Channel.dimension + j
                                    ]
                        elif chn == 13:  # kShaderChannelBlendIndices
                            if not self.m_Skin:
                                self.InitMSkin()
                            for i in range(m_VertexCount):
                                for j in range(m_Channel.dimension):
                                    self.m_Skin[i].boneIndex[j] = componentsIntArray[
                                        i * m_Channel.dimension + j
                                    ]
                    else:
                        if chn == 0:  # kShaderChannelVertex
                            self.m_Vertices = componentsFloatArray
                        elif chn == 1:  # kShaderChannelNormal
                            self.m_Normals = componentsFloatArray
                        elif chn == 2:  # kShaderChannelColor
                            self.m_Colors = componentsFloatArray
                        elif chn == 3:  # kShaderChannelTexCoord0
                            self.m_UV0 = componentsFloatArray
                        elif chn == 4:  # kShaderChannelTexCoord1
                            self.m_UV1 = componentsFloatArray
                        elif chn == 5:
                            if version[0] >= 5:  # kShaderChannelTexCoord2
                                self.m_UV2 = componentsFloatArray
                            else:  # kShaderChannelTangent
                                self.m_Tangents = componentsFloatArray
                        elif chn == 6:  # kShaderChannelTexCoord3
                            self.m_UV3 = componentsFloatArray
                        elif chn == 7:  # kShaderChannelTangent
                            self.m_Tangents = componentsFloatArray

    def DecompressCompressedMesh(self):
        # Vertex
        version = self.version
        m_CompressedMesh = self.m_CompressedMesh
        if m_CompressedMesh.m_Vertices.m_NumItems > 0:
            self.m_VertexCount = int(m_CompressedMesh.m_Vertices.m_NumItems / 3)
            self.m_Vertices = m_CompressedMesh.m_Vertices.UnpackFloats(3, 3 * 4)
        m_VertexCount = self.m_VertexCount
        # UV
        if m_CompressedMesh.m_UV.m_NumItems > 0:  #

            m_UVInfo = m_CompressedMesh.m_UVInfo
            if m_UVInfo != 0:
                kInfoBitsPerUV = 4
                kUVDimensionMask = 3
                kUVChannelExists = 4
                kMaxTexCoordShaderChannels = 8

                uvSrcOffset = 0

                for uv in range(kMaxTexCoordShaderChannels):
                    texCoordBits = m_UVInfo >> (uv * kInfoBitsPerUV)
                    texCoordBits &= (1 << kInfoBitsPerUV) - 1
                    if (texCoordBits & kUVChannelExists) != 0:
                        uvDim = 1 + int(texCoordBits & kUVDimensionMask)
                        m_UV = m_CompressedMesh.m_UV.UnpackFloats(
                            uvDim, uvDim * 4, uvSrcOffset, self.m_VertexCount
                        )
                        self.SetUV(uv, m_UV)
            else:
                self.m_UV0 = m_CompressedMesh.m_UV.UnpackFloats(
                    2, 2 * 4, 0, m_VertexCount
                )
                if m_CompressedMesh.m_UV.m_NumItems >= m_VertexCount * 4:  #
                    self.m_UV1 = m_CompressedMesh.m_UV.UnpackFloats(
                        2, 2 * 4, m_VertexCount * 2, m_VertexCount
                    )

        # BindPose
        if version < (5,):  # 5.0 down
            if m_CompressedMesh.m_BindPoses.m_NumItems > 0:  #
                m_BindPoses_Unpacked = m_CompressedMesh.m_BindPoses.UnpackFloats(
                    16, 4 * 16
                )
                self.m_BindPose = [
                    Matrix4x4(m_BindPoses_Unpacked[i : i + 16])
                    for i in range(0, m_CompressedMesh.m_BindPoses.m_NumItems, 16)
                ]
        # Normal
        if m_CompressedMesh.m_Normals.m_NumItems > 0:
            normalData = m_CompressedMesh.m_Normals.UnpackFloats(2, 4 * 2)
            signs = m_CompressedMesh.m_NormalSigns.UnpackInts()
            self.m_Normals = []  # float[m_CompressedMesh.m_Normals.m_NumItems / 2 * 3]
            for i in range(0, math.ceil(m_CompressedMesh.m_Normals.m_NumItems / 2)):
                x = normalData[i * 2 + 0]
                y = normalData[i * 2 + 1]
                zsqr = 1 - x * x - y * y
                if zsqr >= 0:
                    z = math.sqrt(zsqr)
                else:
                    z = 0
                    normal = Vector3(x, y, z)
                    normal.normalize()
                    x = normal.X
                    y = normal.Y
                    z = normal.Z
                if signs[i] == 0:
                    z = -z
                self.m_Normals.extend([x, y, z])
        # Tangent
        if m_CompressedMesh.m_Tangents.m_NumItems > 0:
            tangentData = m_CompressedMesh.m_Tangents.UnpackFloats(2, 4 * 2)
            signs = m_CompressedMesh.m_TangentSigns.UnpackInts()
            self.m_Tangents = (
                []
            )  # float[m_CompressedMesh.m_Tangents.m_NumItems / 2 * 4]
            for i in range(0, math.ceil(m_CompressedMesh.m_Tangents.m_NumItems / 2)):
                x = tangentData[i * 2 + 0]
                y = tangentData[i * 2 + 1]
                zsqr = 1 - x * x - y * y
                if zsqr >= 0:
                    z = math.sqrt(zsqr)
                else:
                    z = 0
                    vector3f = Vector3(x, y, z)
                    vector3f.normalize()
                    x = vector3f.X
                    y = vector3f.Y
                    z = vector3f.Z
                if signs[i * 2 + 0] == 0:  #
                    z = -z
                w = 1.0 if signs[i * 2 + 1] > 0 else -1.0
                self.m_Tangents.extend([x, y, z, w])

        # FloatColor
        if version >= (5,):  # 5.0 and up
            if m_CompressedMesh.m_FloatColors.m_NumItems > 0:  #
                self.m_Colors = m_CompressedMesh.m_FloatColors.UnpackFloats(1, 4)

        # Skin
        if m_CompressedMesh.m_Weights.m_NumItems > 0:
            weights = m_CompressedMesh.m_Weights.UnpackInts()
            boneIndices = m_CompressedMesh.m_BoneIndices.UnpackInts()
            self.InitMSkin()
            bonePos = 0
            boneIndexPos = 0
            j = 0
            sum = 0

            for i in range(m_CompressedMesh.m_Weights.m_NumItems):
                # read bone index and weight.
                self.m_Skin[bonePos].weight[j] = weights[i] / 31.0
                self.m_Skin[bonePos].boneIndex[j] = boneIndices[boneIndexPos]
                boneIndexPos += 1
                j += 1
                sum += weights[i]

                # the weights add up to one. fill the rest for this vertex with zero, and continue with next one.
                if sum >= 31:  #
                    while j < 4:
                        self.m_Skin[bonePos].weight[j] = 0
                        self.m_Skin[bonePos].boneIndex[j] = 0
                        j += 1

                    bonePos += 1
                    j = 0
                    sum = 0
                # we read three weights, but they don't add up to one. calculate the fourth one, and read
                # missing bone index. continue with next vertex.
                elif j == 3:  #
                    self.m_Skin[bonePos].weight[j] = (31 - sum) / 31.0
                    self.m_Skin[bonePos].boneIndex[j] = boneIndices[boneIndexPos]
                    boneIndexPos += 1
                    bonePos += 1
                    j = 0
                    sum = 0
        # IndexBuffer
        if m_CompressedMesh.m_Triangles.m_NumItems > 0:  #
            self.m_IndexBuffer = m_CompressedMesh.m_Triangles.UnpackInts()
        # Color
        if (
            hasattr(m_CompressedMesh, "m_Colors")
            and m_CompressedMesh.m_Colors.m_NumItems > 0
        ):
            m_CompressedMesh.m_Colors.m_NumItems *= 4
            m_CompressedMesh.m_Colors.m_BitSize /= 4
            tempColors = m_CompressedMesh.m_Colors.UnpackInts()
            self.m_Colors = [color / 255 for color in tempColors]

    def GetTriangles(self):
        m_IndexBuffer = self.m_IndexBuffer
        m_Indices = self.m_Indices

        for m_SubMesh in self.m_SubMeshes:
            firstIndex = m_SubMesh.firstByte // 2
            if not self.m_Use16BitIndices:
                firstIndex //= 2

            indexCount = m_SubMesh.indexCount
            topology = m_SubMesh.topology
            if topology == GfxPrimitiveType.kPrimitiveTriangles:
                m_Indices.extend(
                    m_IndexBuffer[firstIndex : firstIndex + indexCount - indexCount % 3]
                )

            elif (
                self.version[0] < 4
                or topology == GfxPrimitiveType.kPrimitiveTriangleStrip
            ):
                # de-stripify :
                triIndex = 0
                for i in range(indexCount - 2):
                    a, b, c = m_IndexBuffer[firstIndex + i : firstIndex + i + 3]

                    # skip degenerates
                    if a == b or a == c or b == c:
                        continue

                    # do the winding flip-flop of strips :
                    m_Indices.extend([b, a, c] if ((i & 1) == 1) else [a, b, c])
                    triIndex += 3
                # fix indexCount
                m_SubMesh.indexCount = triIndex

            elif topology == GfxPrimitiveType.kPrimitiveQuads:
                for q in range(0, indexCount, 4):
                    m_Indices.extend(
                        [
                            m_IndexBuffer[firstIndex + q],
                            m_IndexBuffer[firstIndex + q + 1],
                            m_IndexBuffer[firstIndex + q + 2],
                            m_IndexBuffer[firstIndex + q],
                            m_IndexBuffer[firstIndex + q + 2],
                            m_IndexBuffer[firstIndex + q + 3],
                        ]
                    )
                # fix indexCount
                m_SubMesh.indexCount = indexCount // 2 * 3

            else:
                raise NotImplementedError(
                    "Failed getting triangles. Submesh topology is lines or points."
                )

    def InitMSkin(self):
        self.m_Skin = [BoneWeights4() for _ in range(self.m_VertexCount)]

    def SetUV(self, uv: int, m_UV):
        if uv == 0:
            self.m_UV0 = m_UV
        elif uv == 1:
            self.m_UV1 = m_UV
        elif uv == 2:
            self.m_UV2 = m_UV
        elif uv == 3:
            self.m_UV3 == m_UV
        elif uv == 4:
            self.m_UV4 == m_UV
        elif uv == 5:
            self.m_UV5 == m_UV
        elif uv == 6:
            self.m_UV6 == m_UV
        elif uv == 7:
            self.m_UV7 == m_UV
        else:
            raise IndexError("Out of Range")

    def GetUV(self, uv: int):
        if uv == 0:
            return self.m_UV0
        elif uv == 1:
            return self.m_UV1
        elif uv == 2:
            return self.m_UV2
        elif uv == 3:
            return self.m_UV3
        elif uv == 4:
            return self.m_UV4
        elif uv == 5:
            return self.m_UV5
        elif uv == 6:
            return self.m_UV6
        elif uv == 7:
            return self.m_UV7
        else:
            raise IndexError("Out of Range")


class MeshHelper:
    @staticmethod
    def GetFormatSize(version, format: int) -> int:
        if version[0] < 2017:
            if format == VertexChannelFormat.kChannelFormatFloat:
                return 4
            elif format == VertexChannelFormat.kChannelFormatFloat16:
                return 2
            elif format == VertexChannelFormat.kChannelFormatColor:  # in 4.x is size 4
                return 1
            elif format == VertexChannelFormat.kChannelFormatByte:
                return 1
            elif format == VertexChannelFormat.kChannelFormatUInt32:  # in 5.x
                return 4
        elif version[0] < 2019:
            if format == VertexFormat.kVertexFormatFloat:
                return 4
            elif format == VertexFormat.kVertexFormatFloat16:
                return 2
            elif format == VertexFormat.kVertexFormatColor:
                return 1
            elif format == VertexFormat.kVertexFormatUNorm8:
                return 1
            elif format == VertexFormat.kVertexFormatSNorm8:
                return 1
            elif format == VertexFormat.kVertexFormatUNorm16:
                return 2
            elif format == VertexFormat.kVertexFormatSNorm16:
                return 2
            elif format == VertexFormat.kVertexFormatUInt8:
                return 1
            elif format == VertexFormat.kVertexFormatSInt8:
                return 1
            elif format == VertexFormat.kVertexFormatUInt16:
                return 2
            elif format == VertexFormat.kVertexFormatSInt16:
                return 2
            elif format == VertexFormat.kVertexFormatUInt32:
                return 4
            elif format == VertexFormat.kVertexFormatSInt32:
                return 4
        else:
            if format == VertexFormatV2019.kVertexFormatFloat:
                return 4
            elif format == VertexFormatV2019.kVertexFormatFloat16:
                return 2
            elif format == VertexFormatV2019.kVertexFormatUNorm8:
                return 1
            elif format == VertexFormatV2019.kVertexFormatSNorm8:
                return 1
            elif format == VertexFormatV2019.kVertexFormatUNorm16:
                return 2
            elif format == VertexFormatV2019.kVertexFormatSNorm16:
                return 2
            elif format == VertexFormatV2019.kVertexFormatUInt8:
                return 1
            elif format == VertexFormatV2019.kVertexFormatSInt8:
                return 1
            elif format == VertexFormatV2019.kVertexFormatUInt16:
                return 2
            elif format == VertexFormatV2019.kVertexFormatSInt16:
                return 2
            elif format == VertexFormatV2019.kVertexFormatUInt32:
                return 4
            elif format == VertexFormatV2019.kVertexFormatSInt32:
                return 4
        raise ValueError(format)

    @staticmethod
    def IsIntFormat(version, format: int) -> bool:
        if version[0] < 2017:
            return format == 4
        elif version[0] < 2019:
            return format >= 7
        else:
            return format >= 6

    @staticmethod
    def BytesToFloatArray(inputBytes, size):
        return [
            inputBytes[i] / 255.0
            if size == 1
            else struct.unpack(">e", inputBytes[i * 2 : i * 2 + 2])[0]
            if size == 2
            else struct.unpack(">f", inputBytes[i * 4 : i * 4 + 4])[0]
            if size == 4
            else 0
            for i in range(len(inputBytes) // size)
        ]

    @staticmethod
    def BytesToIntArray(inputBytes, size):
        return [
            int.from_bytes(inputBytes[i : i + size], byteorder="big", signed=True)
            for i in range(0, len(inputBytes), size)
        ]


class VertexChannelFormat(IntEnum):
    kChannelFormatFloat = 0
    kChannelFormatFloat16 = 1
    kChannelFormatColor = 2
    kChannelFormatByte = 3
    kChannelFormatUInt32 = 4


class VertexFormat(IntEnum):
    kVertexFormatFloat = 0
    kVertexFormatFloat16 = 1
    kVertexFormatColor = 2
    kVertexFormatUNorm8 = 3
    kVertexFormatSNorm8 = 4
    kVertexFormatUNorm16 = 5
    kVertexFormatSNorm16 = 6
    kVertexFormatUInt8 = 7
    kVertexFormatSInt8 = 8
    kVertexFormatUInt16 = 9
    kVertexFormatSInt16 = 10
    kVertexFormatUInt32 = 11
    kVertexFormatSInt32 = 12


class VertexFormatV2019(IntEnum):
    kVertexFormatFloat = 0
    kVertexFormatFloat16 = 1
    kVertexFormatUNorm8 = 2
    kVertexFormatSNorm8 = 3
    kVertexFormatUNorm16 = 4
    kVertexFormatSNorm16 = 5
    kVertexFormatUInt8 = 6
    kVertexFormatSInt8 = 7
    kVertexFormatUInt16 = 8
    kVertexFormatSInt16 = 9
    kVertexFormatUInt32 = 10
    kVertexFormatSInt32 = 11



from .Behaviour import Behaviour
from .PPtr import PPtr


class Animator(Behaviour):
    def __init__(self, reader):
        super().__init__(reader=reader)

        self.m_Avatar = PPtr(reader)  # Avatar
        self.m_Controller = PPtr(reader)  # RuntimeAnimatorController
        self.m_CullingMode = reader.read_int()
        version = self.version

        if version >= (4, 5):  # 4.5 and up
            self.m_UpdateMode = reader.read_int()

        self.m_ApplyRootMotion = reader.read_boolean()
        if (4, 5) < version[2:] <= (5, 0):  # 4.5 and up - 5.0 down
            reader.align_stream()

        if version >= (5,):  # 5.0 and up
            self.m_LinearVelocityBlending = reader.read_boolean()
            reader.align_stream()

        if version[2:] < (4, 5):  # 4.5 down
            self.m_AnimatePhysics = reader.read_boolean()

        if version >= (4, 3):  # 4.3 and up
            self.m_HasTransformHierarchy = reader.read_boolean()

        if version >= (4, 5):  # 4.5 and up
            self.m_AllowConstantClipSamplingOptimization = reader.read_boolean()

        if (4,) < version[:1] < (2018,):  # 5.0 and up - 2018 down
            reader.align_stream()

        if version >= (2018,):  # 2018 and up
            self.m_KeepAnimatorControllerStateOnDisable = reader.read_boolean()
            reader.align_stream()



from .Object import Object
from .PPtr import PPtr, save_ptr
from ..enums import BuildTarget
from ..streams import EndianBinaryWriter


class EditorExtension(Object):
    def __init__(self, reader):
        super().__init__(reader=reader)
        if self.platform == BuildTarget.NoTarget:
            self.m_PrefabParentObject = PPtr(reader)
            self.m_PrefabInternal = PPtr(reader)

    def save(self, writer: EndianBinaryWriter = None):
        if not writer:
            writer = EndianBinaryWriter(endian = self.reader.endian)
        super().save(writer, intern_call=True)
        if self.platform == BuildTarget.NoTarget:
            save_ptr(self.m_PrefabParentObject, writer)
            save_ptr(self.m_PrefabInternal, writer)



import os, glob
from ..streams import EndianBinaryReader
from ..files import File


def get_resource_data(*args):
    """
    Input:
    Option 1:
        0 - path - file path
        1 - assets_file - SerializedFile
        2 - offset -
        3 - size -
    Option 2:
        0 - reader - EndianBinaryReader
        1 - offset -
        2 - size -

    -> -2 = offset, -1 = size
    """
    if len(args) == 4:
        reader = search_resource(res_path=args[0], assets_file=args[1])
    elif len(args) == 3:
        reader = args[0]
    else:
        raise TypeError(f"3 or 4 arguments required, but only {len(args)} given")

    reader.Position = args[-2]
    return reader.read_bytes(args[-1])


def search_resource(res_path, assets_file):
    # try to find the resource in the Unity packages
    base_name = os.path.basename(res_path)
    if os.path.splitext(base_name)[1] == ".resource":
        base_name2 = base_name.replace('.resource', '.assets.resS')
    else:
        base_name2 = base_name.replace('.assets.resS', '.resource')

    for p in [res_path, base_name, base_name2]:
        reader = assets_file.parent.files.get(p)
        if reader:
            if isinstance(reader, File.File):
                # in case the import helper accidentally detected a resource file as something else
                reader = reader.reader
            return reader

    # try to find it in the dir environment
    c = assets_file
    path = getattr(assets_file, "path", None)
    while not path:
        c = getattr(c,"parent",None)
        if c == None:
            raise FileNotFoundError(
                f"Can't find the resource file {res_path}"
            )
        path = getattr(c, "path", None)
    current_directory = path
    resource_file_path = os.path.join(current_directory, *res_path.split("/"))
    if not os.path.isfile(resource_file_path):
        resource_file_path = search_resource_file(current_directory, base_name)
    if not os.path.isfile(resource_file_path):
        resource_file_path = search_resource_file(current_directory, base_name.replace('.assets.resS', '.resource'))

    if os.path.isfile(resource_file_path):
        return EndianBinaryReader(open(resource_file_path, "rb"))
    else:
        raise FileNotFoundError(
            f"Can't find the resource file {res_path}"
        )


def search_resource_file(path, name):
    #print("real file", path, name)
    files = glob.glob(os.path.join(path, "**", name), recursive=True)
    return files[0] if len(files) else ""




class Quaternion:
    X: float
    Y: float
    Z: float
    W: float
    
    def __init__(self, x: float = 0.0, y: float = 0.0, z: float = 0.0, w: float = 0.0):
        self._data = [0.0] * 4
        self.X = x
        self.Y = y
        self.Z = z
        self.W = w
    
    @property
    def X(self) -> float:
        return self._data[0]
    
    @X.setter
    def X(self, value: float):
        self._data[0] = value
    
    @property
    def Y(self) -> float:
        return self._data[1]
    
    @Y.setter
    def Y(self, value: float):
        self._data[1] = value
    
    @property
    def Z(self) -> float:
        return self._data[2]
    
    @Z.setter
    def Z(self, value: float):
        self._data[2] = value
    
    @property
    def W(self) -> float:
        return self._data[3]
    
    @W.setter
    def W(self, value: float):
        self._data[3] = value
    
    def __getitem__(self, value):
        return self._data[value]
    
    def __setitem__(self, index, value):
        self._data[index] = value



from dataclasses import dataclass
from math import sqrt


kEpsilon = 0.00001


@dataclass
class Vector3:
    X: float = 0.0
    Y: float = 0.0
    Z: float = 0.0

    def __init__(self, *args):
        if len(args) == 3 or len(args) == 1 and isinstance(args[0], (tuple, list)):
            self.X, self.Y, self.Z = args
        elif len(args) == 1:
            # dirty patch for Vector4
            self.__dict__ = args[0].__dict__

    def __getitem__(self, index):
        return (self.X, self.Y, self.Z)[index]

    def __setitem__(self, index, value):
        if index == 0:
            self.X = value
        elif index == 1:
            self.Y = value
        elif index == 2:
            self.Z = value
        else:
            raise IndexError("Index out of range")

    def __hash__(self):
        return self.X.__hash__() ^ (self.Y.__hash__() << 2) ^ (self.Z.__hash__() >> 2)

    def __eq__(self, other):
        if isinstance(other, Vector3):
            return self.X == other.X and self.Y == other.Y and self.Z == other.Z
        else:
            return False

    def normalize(self):
        length = self.length()
        if length > kEpsilon:
            invNorm = 1.0 / length
            self.X *= invNorm
            self.Y *= invNorm
            self.Z *= invNorm
        else:
            X = 0
            Y = 0
            Z = 0

    def Normalize(self):
        self.normalize()

    def length(self):
        return sqrt(self.LengthSquared())

    def Length(self):
        return self.length()

    def LengthSquared(self):
        return self.X ** 2 + self.Y ** 2 + self.Y ** 2

    @staticmethod
    def Zero():
        return Vector3(0, 0, 0)

    @staticmethod
    def One():
        return Vector3(1, 1, 1)

    def __add__(a, b):
        return Vector3(a.X + b.X, a.Y + b.Y, a.Z + b.Z)

    def __sub__(a, b):
        return Vector3(a.X - b.X, a.Y - b.Y, a.Z - b.Z)

    def __mul__(a, d):
        return Vector3(a.X * d, a.Y * d, a.Z * d)

    def __div__(a, d):
        return Vector3(a.X / d, a.Y / d, a.Z / d)

    def __eq__(lhs, rhs):
        return (lhs - rhs).LengthSquared() < kEpsilon

    def __ne__(lhs, rhs):
        return not (lhs == rhs)

    def Vector2(self):
        from .Vector2 import Vector2
        return Vector2(self.X, self.Y)

    def Vector4(self):
        from .Vector4 import Vector4
        return Vector4(self.X, self.Y, self.Z, 0.0)



from enum import IntEnum

# based on https://stackoverflow.com/a/57179436


class ExtendableEnum(IntEnum):
    @classmethod
    def _missing_(cls, value):
        if isinstance(value, int):
            pseudo_member = cls._value2member_map_.get(value, None)
            if pseudo_member is None:
                new_member = int.__new__(cls, value)
                # I expect a name attribute to hold a string, hence str(value)
                # However, new_member._name_ = value works, too
                new_member._name_ = f"Unknown ({value})"
                new_member._value_ = value
                pseudo_member = cls._value2member_map_.setdefault(value, new_member)
            return pseudo_member
        return None  # will raise the ValueError in Enum.__new__



from .PPtr import PPtr
from ..enums import BuildTarget
from ..helpers import TypeTreeHelper
from ..streams import EndianBinaryWriter
from ..files import ObjectReader
import types
from ..exceptions import TypeTreeError as TypeTreeError


class Object(object):
    type_tree: dict

    def __init__(self, reader: ObjectReader):
        self.reader = reader
        self.assets_file = reader.assets_file
        self.type = reader.type
        self.path_id = reader.path_id
        self.version = reader.version
        self.build_type = reader.build_type
        self.platform = reader.platform
        self.serialized_type = reader.serialized_type
        self.byte_size = reader.byte_size
        self.assets_file = reader.assets_file

        if self.platform == BuildTarget.NoTarget:
            self._object_hide_flags = reader.read_u_int()

        self.container = (
            self.assets_file._container[self.path_id]
            if self.path_id in self.assets_file._container
            else None
        )

        self.reader.reset()
        if type(self) == Object:
            self.read_typetree()

    def has_struct_member(self, name: str) -> bool:
        return self.serialized_type.m_Nodes and any(
            [x.name == name for x in self.serialized_type.m_Nodes]
        )

    def dump_typetree(self, nodes: list = None) -> str:
        return self.reader.dump_typetree(nodes=nodes)

    def dump_typetree_structure(self) -> str:
        return self.reader.dump_typetree_structure()

    def read_typetree(self, nodes: list = None) -> dict:
        try:
            tree = self.reader.read_typetree(nodes)
        except TypeTreeError as e:
            print("Failed to read TypeTree:\n", e)
            return {}
        self.type_tree = NodeHelper(tree, self.assets_file)
        return tree

    def save_typetree(self, nodes: list = None, writer: EndianBinaryWriter = None):
        if not writer:
            writer = EndianBinaryWriter(endian=self.reader.endian)

        def class_to_dict(value):
            if isinstance(value, list):
                return [class_to_dict(val) for val in value]
            elif isinstance(value, dict):
                return {key: class_to_dict(val) for key, val in value.items()}
            elif hasattr(value, "__dict__"):
                if isinstance(value, PPtr):
                    return {"m_PathID": value.path_id, "m_FileID": value.file_id}
                return {
                    key: class_to_dict(val)
                    for key, val in value.__dict__.items()
                    if not isinstance(value, (types.FunctionType, types.MethodType))
                    and not key in ["type_tree", "assets_file"]
                }
            else:
                return value

        obj = class_to_dict(self if not self.type_tree else self.type_tree)
        return self.reader.save_typetree(obj, nodes, writer)

    def get_raw_data(self) -> bytes:
        return self.reader.get_raw_data()

    def set_raw_data(self, data):
        self.reader.set_raw_data(data)

    def save(self, writer: EndianBinaryWriter = None, intern_call=False):
        if not writer:
            writer = EndianBinaryWriter(endian=self.reader.endian)
        if intern_call:
            if self.platform == BuildTarget.NoTarget:
                writer.write_u_int(self._object_hide_flags)
        elif self.serialized_type.nodes:
            # save for objects WITHOUT specific save function
            # so we have to use the typetree if it exists
            self.save_typetree()
        else:
            raise NotImplementedError(
                "There is no save function for this obj.type nor has it any typetree nodes that could be used."
            )

    def _save(self, writer):
        # the reader is actually an ObjectReader,
        # the data value is written back into the asset
        self.reader.data = writer.bytes

    def __getattr__(self, name):
        """
        If item not found in __dict__, read type_tree and check if it is in there.
        """
        if name == "type_tree" or self.type_tree == None:
            old_pos = self.reader.Position
            self.read_typetree()
            self.reader.Position = old_pos
            if name == "type_tree":
                return self.type_tree
        elif name == "read":
            return lambda: self
        return getattr(self.type_tree, name)

    def get(self, key, default=None):
        return getattr(self, key, default)

    def __repr__(self):
        return "<%s %s>" % (self.__class__.__name__, self.name)

    def __hash__(self):
        return hash(self.path_id)

    def __eq__(self, other):
        return self.path_id == other.path_id


class NodeHelper:
    def __init__(self, data, assets_file):
        if "m_PathID" in data and "m_FileID" in data:
            # used to make pointers directly useable
            self.path_id = data["m_PathID"]
            self.file_id = data["m_FileID"]
            self.index = data.get("m_Index", -2)
            self.assets_file = assets_file
            self._obj = None
            self.__class__ = PPtr
        else:
            self.__dict__ = {
                key: NodeHelper(val, assets_file) for key, val in data.items()
            }

    def __new__(cls, data, assets_file):
        if isinstance(data, dict):
            return super(NodeHelper, cls).__new__(cls)
        elif isinstance(data, list):
            return [NodeHelper(x, assets_file) for x in data]
        return data

    def __getitem__(self, item):
        return getattr(self, item)

    def to_dict(self):
        def dump(val):
            return (
                val.to_dict()
                if isinstance(val, NodeHelper)
                else [dump(item) for item in val]
                if isinstance(val, list)
                else {"m_PathID": val.path_id, "m_FileID": val.file_id}
                if isinstance(val, PPtr)
                else [x for x in val]
                if isinstance(val, (bytearray, bytes))
                else val
            )

        return {key: dump(val) for key, val in self.__dict__.items()}

    def items(self):
        return self.__dict__.items()

    def values(self):
        return self.__dict__.values()

    def keys(self):
        return self.__dict__.keys()

    def __repr__(self):
        return "<NodeHelper - %s>" % self.__dict__.__repr__()



from . import File
from ..helpers import CompressionHelper
from ..streams import EndianBinaryReader, EndianBinaryWriter

from collections import namedtuple

BlockInfo = namedtuple("BlockInfo", "uncompressedSize compressedSize flags")
DirectoryInfoFS = namedtuple("DirectoryInfoFS", "offset size flags path")


class BundleFile(File.File):
    format: int
    is_changed: bool
    signature: str
    version_engine: str
    version_player: str

    def __init__(self, reader: EndianBinaryReader, parent: File, name: str=None):
        super().__init__(parent=parent, name=name)
        signature = self.signature = reader.read_string_to_null()
        self.version = reader.read_u_int()
        self.version_player = reader.read_string_to_null()
        self.version_engine = reader.read_string_to_null()

        if signature == "UnityArchive":
            raise NotImplemented("BundleFile - UnityArchive")
        elif signature in ["UnityWeb", "UnityRaw"]:
            m_DirectoryInfo, blocksReader = self.read_web_raw(reader)
        elif signature == "UnityFS":
            m_DirectoryInfo, blocksReader = self.read_fs(reader)
        else:
            raise NotImplemented(f"Unknown Bundle signature: {signature}")

        self.read_files(blocksReader, m_DirectoryInfo)

    def read_web_raw(self, reader: EndianBinaryReader):
        # def read_header_and_blocks_info(self, reader:EndianBinaryReader):
        version = self.version
        if version >= 4:
            _hash = reader.read_bytes(16)
            crc = reader.read_u_int()

        minimumStreamedBytes = reader.read_u_int()
        headerSize = reader.read_u_int()
        numberOfLevelsToDownloadBeforeStreaming = reader.read_u_int()
        levelCount = reader.read_int()
        reader.Position += 4 * 2 * (levelCount - 1)

        compressedSize = reader.read_u_int()
        uncompressedSize = reader.read_u_int()

        if version >= 2:
            completeFileSize = reader.read_u_int()

        if version >= 3:
            fileInfoHeaderSize = reader.read_u_int()

        reader.Position = headerSize

        uncompressedBytes = CompressionHelper.decompress_lzma(
            reader.read_bytes(compressedSize)
        )

        blocksReader = EndianBinaryReader(uncompressedBytes, offset=headerSize)
        nodesCount = blocksReader.read_int()
        m_DirectoryInfo = [
            File.DirectoryInfo(
                blocksReader.read_string_to_null(),  # path
                blocksReader.read_u_int(),  # offset
                blocksReader.read_u_int(),  # size
            )
            for _ in range(nodesCount)
        ]

        return m_DirectoryInfo, blocksReader

    def read_fs(self, reader: EndianBinaryReader):
        version = self.version
        size = reader.read_long()

        # header
        compressedSize = reader.read_u_int()
        uncompressedSize = reader.read_u_int()
        self._data_flags = reader.read_u_int()

        if self.version >= 7:
            reader.align_stream(16)

        _position = reader.Position
        if self._data_flags & 0x80 != 0:  # kArchiveBlocksInfoAtTheEnd
            reader.Position = reader.Length - compressedSize
            blocksInfoBytes = reader.read_bytes(compressedSize)
            reader.Position = _position
        else:  # 0x40 kArchiveBlocksAndDirectoryInfoCombined
            blocksInfoBytes = reader.read_bytes(compressedSize)

        switch = self._data_flags & 0x3F

        if switch == 1:  # LZMA
            blocksInfoBytes = CompressionHelper.decompress_lzma(blocksInfoBytes)
        elif switch in [2, 3]:  # LZ4, LZ4HC
            blocksInfoBytes = CompressionHelper.decompress_lz4(
                blocksInfoBytes, uncompressedSize
            )
        # elif switch == 4: #LZHAM:

        blocksInfoReader = EndianBinaryReader(blocksInfoBytes, offset=_position)

        uncompressedDataHash = blocksInfoReader.read_bytes(16)
        blocksInfoCount = blocksInfoReader.read_int()

        m_BlocksInfo = [
            BlockInfo(
                blocksInfoReader.read_u_int(),  # uncompressedSize
                blocksInfoReader.read_u_int(),  # compressedSize
                blocksInfoReader.read_u_short(),  # flags
            )
            for _ in range(blocksInfoCount)
        ]

        nodesCount = blocksInfoReader.read_int()
        m_DirectoryInfo = [
            DirectoryInfoFS(
                blocksInfoReader.read_long(),  # offset
                blocksInfoReader.read_long(),  # size
                blocksInfoReader.read_u_int(),  # flags
                blocksInfoReader.read_string_to_null(),  # path
            )
            for _ in range(nodesCount)
        ]

        # def read_blocks(reader : EndianBinaryReader, blocksStream):
        def decompress_block(blockInfo):
            switch = blockInfo.flags & 0x3F  # kStorageBlockCompressionTypeMask
            if switch == 1:  # LZMA
                return CompressionHelper.decompress_lzma(
                    reader.read_bytes(blockInfo.compressedSize)
                )
            elif switch in [2, 3]:  # LZ4, LZ4HC
                return CompressionHelper.decompress_lz4(
                    reader.read_bytes(blockInfo.compressedSize),
                    blockInfo.uncompressedSize,
                )
            # elif switch == 4: #LZHAM:
            else:  # no compression
                return reader.read_bytes(blockInfo.uncompressedSize)

        if m_BlocksInfo:
            self._block_info_flags = m_BlocksInfo[0].flags

        blocksReader = EndianBinaryReader(
            b"".join(decompress_block(blockInfo) for blockInfo in m_BlocksInfo),
            offset=(blocksInfoReader.real_offset())
        )

        return m_DirectoryInfo, blocksReader


    def save(self, packer=None):
        """
            Rewrites the BundleFile and returns it as bytes object.

            packer:
                can be either one of the following strings
                or tuple consisting of (block_info_flag, data_flag)
                allowed strings:
                    none - no compression, default, safest bet
                    lz4 - lz4 compression
                    original - uses the original flags
        """
        # file_header
        #     signature    (string_to_null)
        #     format        (int)
        #     version_player    (string_to_null)
        #     version_engine    (string_to_null)
        writer = EndianBinaryWriter()

        writer.write_string_to_null(self.signature)
        writer.write_u_int(self.version)
        writer.write_string_to_null(self.version_player)
        writer.write_string_to_null(self.version_engine)

        if self.signature == "UnityArchive":
            raise NotImplemented("BundleFile - UnityArchive")
        elif self.signature in ["UnityWeb", "UnityRaw"]:
            raise NotImplemented("Saving Unity Web and Raw bundles isn't supported yet")
            # self.save_web_raw(writer)
        elif self.signature == "UnityFS":
            if not packer or packer == "none":
                self.save_fs(writer, 64, 64)
            elif packer == "original":
                self.save_fs(writer, block_info_flag=self._block_info_flags, data_flag=self._data_flags)
            elif packer == "lz4":
                self.save_fs(writer, block_info_flag=194, data_flag=2)
            elif isinstance(packer, tuple):
                self.save_fs(writer, *packer)
            else:
                raise NotImplemented("UnityFS - Packer:", packer)
        return writer.bytes

    def save_fs(self, writer: EndianBinaryWriter, block_info_flag : int, data_flag : int):
        # header
        # compressed blockinfo (block details & directionary)
        # compressed assets

        # 0b1000000 / 0b11000000 | 64 / 192 - uncompressed
        # 0b11000010 | 194 - lz4
        # block_info_flag

        # 0 / 0b1000000 | 0 / 64 - uncompressed
        # 0b1   | 1 - lzma
        # 0b10  | 2 - lz4
        # 0b11  | 3 - lz4hc [not implemented]
        # 0b100 | 4 - lzham [not implemented]
        # data_flag

        # header:
        #     bundle_size        (long)
        #     compressed_size    (int)
        #     uncompressed_size    (int)
        #     flag                (int)
        #     ?padding?            (bool)
        #   This will be written at the end,
        #   because the size can only be calculated after the data compression,

        # block_info:
        #     *flag & 0x80 ? at the end : right after header
        #     *decompression via flag & 0x3F
        #     *read compressed_size -> uncompressed_size
        #     0x10 offset
        #     *read blocks infos of the data stream
        #     count            (int)
        #     (
        #         uncompressed_size(uint)
        #         compressed_size (uint)
        #         flag(short)
        #     )
        #     *decompression via info.flag & 0x3F

        #     *afterwards the file positions
        #     file_count        (int)
        #     (
        #         offset    (long)
        #         size        (long)
        #         flag        (int)
        #         name        (string_to_null)
        #     )

        # file list & file data
        # prep nodes and build up block data
        data_writer = EndianBinaryWriter()
        files = [
            (
                name,
                f.flags,
                data_writer.write_bytes(
                    f.bytes if isinstance(f, (EndianBinaryReader, EndianBinaryWriter)) else f.save()
                ),
            )
            for name, f in self.files.items()
        ]

        file_data = data_writer.bytes
        data_writer.dispose()
        uncompressed_data_size = len(file_data)

        # compress the data
        switch = data_flag & 0x3F
        if switch == 1:  # LZMA
            file_data = CompressionHelper.compress_lzma(file_data)
        elif switch in [2, 3]:  # LZ4, LZ4HC
            file_data = CompressionHelper.compress_lz4(file_data)
        elif switch == 4:  # LZHAM
            raise NotImplementedError
        # else no compression - data stays the same
        compressed_data_size = len(file_data)

        # write the block_info
        # uncompressedDataHash
        block_writer = EndianBinaryWriter(b"\x00" * 0x10)
        # data block info
        # block count
        block_writer.write_int(1)
        # uncompressed size
        block_writer.write_u_int(uncompressed_data_size)
        # compressed size
        block_writer.write_u_int(compressed_data_size)
        # flag
        block_writer.write_u_short(data_flag)

        # file block info
        # file count
        block_writer.write_int(len(files))
        offset = 0
        for f_name, f_flag, f_len in files:
            # offset
            block_writer.write_long(offset)
            # size
            block_writer.write_long(f_len)
            offset += f_len
            # flag
            block_writer.write_u_int(f_flag)
            # name
            block_writer.write_string_to_null(f_name)

        # compress the block data
        block_data = block_writer.bytes
        block_writer.dispose()

        uncompressed_block_data_size = len(block_data)

        switch = block_info_flag & 0x3F
        if switch == 1:  # LZMA
            block_data = CompressionHelper.compress_lzma(block_data)
        elif switch in [2, 3]:  # LZ4, LZ4HC
            block_data = CompressionHelper.compress_lz4(block_data)
        elif switch == 4:  # LZHAM
            raise NotImplementedError

        compressed_block_data_size = len(block_data)

        # write the header info
        ## file size
        writer.write_long(
            writer.Length
            + 8
            + 4
            + 4
            + 4
            + (8 if self.version >= 7 else 0)
            + compressed_block_data_size
            + compressed_data_size
        )
        # compressed blockInfoBytes size
        writer.write_u_int(compressed_block_data_size)
        # uncompressed size
        writer.write_u_int(uncompressed_block_data_size)
        # compression flag
        writer.write_u_int(block_info_flag)

        if self.version >= 7:
            # UnityFS\x00 - 8
            # size 8
            # comp sizes 4+4
            # flag 4
            # sum : 28 -> +8 alignment
            writer.align_stream(16)

        if (block_info_flag & 0x80) != 0:  # at end of file
            writer.write(file_data)
            writer.write(block_data)
        else:
            writer.write(block_data)
            writer.write(file_data)



from enum import IntEnum

from .Mesh import BoneWeights4, SubMesh, VertexData
from .NamedObject import NamedObject
from .PPtr import PPtr
from ..export import SpriteHelper
from ..enums import SpriteMeshType, SpritePackingMode, SpritePackingRotation
from ..streams import EndianBinaryWriter, EndianBinaryReader


class Sprite(NamedObject):
    @property
    def image(self):
        return SpriteHelper.get_image_from_sprite(self)

    def __init__(self, reader):
        super().__init__(reader=reader)
        version = self.version

        self.m_Rect = reader.read_rectangle_f()
        self.m_Offset = reader.read_vector2()
        if version >= (4, 5):  # 4.5 and up
            self.m_Border = reader.read_vector4()

        self.m_PixelsToUnits = reader.read_float()
        if version >= (5, 4, 2) or (
            version >= (5, 4, 1, 3) and self.build_type.IsPatch
        ):  # 5.4.1p3 and up
            self.m_Pivot = reader.read_vector2()

        self.m_Extrude = reader.read_u_int()
        if version >= (5, 3):  # 5.3 and up
            self.m_IsPolygon = reader.read_boolean()
            reader.align_stream()

        if version >= (2017,):  # 2017 and up
            first = reader.read_bytes(16)  # GUID
            second = reader.read_long()
            self.m_RenderDataKey = (first, second)
            self.m_AtlasTags = reader.read_string_array()
            self.m_SpriteAtlas = PPtr(reader)  # SpriteAtlas

        self.m_RD = SpriteRenderData(reader)

        if version >= (2017,):  # 2017 and up
            m_PhysicsShapeSize = reader.read_int()
            self.m_PhysicsShape = [
                reader.read_vector2_array() for _ in range(m_PhysicsShapeSize)
            ]

        if version >= (2018,):  # 2018 and up
            m_BonesSize = reader.read_int()
            # TODO: might occur in earlier 2020 versions - 2020.3.13 reported
            if version >= (
                2020,
                3,
            ):
                self.m_Bones = [SpriteBone() for _ in range(m_BonesSize)]
            else:
                self.m_Bones = [reader.read_vector2_array() for _ in range(m_BonesSize)]

    def save(self, writer: EndianBinaryWriter = None):
        if writer is None:
            writer = EndianBinaryWriter(endian=self.reader.endian)
        version = self.version

        super().save(writer)
        writer.write_rectangle_f(self.m_Rect)
        writer.write_vector2(self.m_Offset)
        if version >= (4, 5):  # 4.5 and up
            writer.write_vector4(self.m_Border)

        writer.write_float(self.m_PixelsToUnits)
        if version >= (5, 4, 2) or (
            version >= (5, 4, 1, 3) and self.build_type.IsPatch
        ):  # 5.4.1p3 and up
            writer.write_vector2(self.m_Pivot)

        writer.write_u_int(self.m_Extrude)
        if version >= (5, 3):  # 5.3 and up
            writer.write_boolean(self.m_IsPolygon)
            writer.align_stream()

        if version >= (2017,):  # 2017 and up
            writer.write_bytes(self.m_RenderDataKey[0])  # GUID
            writer.write_long(self.m_RenderDataKey[1])
            writer.write_string_array(self.m_AtlasTags)
            self.m_SpriteAtlas.save(writer)  # SpriteAtlas

        self.m_RD.save(writer, version)

        if version >= (2017,):  # 2017 and up
            writer.write_int(len(self.m_PhysicsShape))
            for phys in self.m_PhysicsShape:
                writer.write_vector2_array(phys)

        if version >= (2018,):  # 2018 and up
            writer.write_int(len(self.m_Bones))
            if version >= (2020, 3):
                for bone in self.m_Bones:
                    bone.save(writer)
            else:
                for bone in self.m_Bones:
                    writer.write_vector2_array(bone)

        self.set_raw_data(writer.bytes)


class SecondarySpriteTexture:
    def __init__(self, reader):
        self.texture = PPtr(reader)  # Texture2D
        self.name = reader.read_string_to_null()

    def save(self, writer):
        self.texture.save(writer)
        writer.write_string_to_null(self.name)


class SpriteSettings:
    def __init__(self, reader):
        self.settingsRaw = reader.read_u_int()
        self.packed = self.settingsRaw & 1  # 1
        self.packingMode = SpritePackingMode((self.settingsRaw >> 1) & 1)  # 1
        self.packingRotation = SpritePackingRotation((self.settingsRaw >> 2) & 0xF)  # 4
        self.meshType = SpriteMeshType((self.settingsRaw >> 6) & 1)  # 1
        # rest of the bits are reserved

    def save(self, writer):
        writer.write_u_int(self.settingsRaw)


class SpriteVertex:
    def __init__(self, reader):
        version = reader.version

        self.pos = reader.read_vector3()
        if version[:2] <= (4, 3):  # 4.3 and down
            self.uv = reader.read_vector2()

    def save(self, writer, version):
        writer.write_vector3(self.pos)
        if version[:2] <= (4, 3):  # 4.3 and down
            writer.write__vector2(self.uv)


class SpriteRenderData:
    def __init__(self, reader):
        version = reader.version

        self.texture = PPtr(reader)  # Texture2D
        if version >= (5, 2):  # 5.2 and up
            self.alphaTexture = PPtr(reader)  # Texture2D

        if version >= (2019,):  # 2019 and up
            secondaryTexturesSize = reader.read_int()
            self.secondaryTextures = [
                SecondarySpriteTexture(reader) for _ in range(secondaryTexturesSize)
            ]

        if version >= (5, 6):  # 5.6 and up
            SubMeshesSize = reader.read_int()
            self.m_SubMeshes = [SubMesh(reader) for _ in range(SubMeshesSize)]
            IndexBufferSize = reader.read_int()
            self.m_IndexBuffer = reader.read_bytes(IndexBufferSize)
            reader.align_stream()
            self.m_VertexData = VertexData(reader)
        else:
            verticesSize = reader.read_int()
            self.vertices = [SpriteVertex(reader) for _ in range(verticesSize)]
            self.indices = reader.read_u_short_array()
            reader.align_stream()

        if version >= (2018,):  # 2018 and up
            self.m_Bindpose = reader.read_matrix_array()
            if version < (2018, 2):  # 2018.2 down
                self.m_SourceSkinSize = reader.read_int()
                self.m_SourceSkin = [BoneWeights4(reader)]

        self.textureRect = reader.read_rectangle_f()
        self.textureRectOffset = reader.read_vector2()
        if version >= (5, 6):  # 5.6 and up
            self.atlasRectOffset = reader.read_vector2()

        self.settingsRaw = SpriteSettings(reader)
        if version >= (4, 5):  # 4.5 and up
            self.uvTransform = reader.read_vector4()

        if version >= (2017,):  # 2017 and up
            self.downscaleMultiplier = reader.read_float()

    def save(self, writer, version):
        self.texture.save(writer)  # Texture2D
        if version >= (5, 2):  # 5.2 and up
            self.alphaTexture.save(writer)  # Texture2D

        if version >= (2019,):  # 2019 and up
            writer.write_int(len(self.secondaryTextures))
            for tex in self.secondaryTextures:
                tex.save(writer)

        if version >= (5, 6):  # 5.6 and up
            writer.write_int(len(self.m_SubMeshes))
            for mesh in self.m_SubMeshes:
                mesh.save(writer, version)
            writer.write_int(len(self.m_IndexBuffer))
            writer.write_bytes(self.m_IndexBuffer)
            writer.align_stream()
            self.m_VertexData.save(writer, version)
        else:
            writer.write_int(len(self.vertices))
            for vertex in self.vertices:
                vertex.save(writer, version)
            writer.write_u_short_array(self.indices)
            writer.align_stream()

        if version >= (2018,):  # 2018 and up
            writer.write_matrix_array(self.m_Bindpose)
            if version < (2018, 2):  # 2018.2 down
                writer.write_int(self.m_SourceSkinSize)
                self.m_SourceSkin[0].save(writer)

        writer.write_rectangle_f(self.textureRect)
        writer.write_vector2(self.textureRectOffset)
        if version >= (5, 6):  # 5.6 and up
            writer.write_vector2(self.atlasRectOffset)

        self.settingsRaw.save(writer)
        if version >= (4, 5):  # 4.5 and up
            writer.write_vector4(self.uvTransform)

        if version >= (2017,):  # 2017 and up
            writer.write_float(self.downscaleMultiplier)


class SpriteBone:
    name: str
    position: tuple
    rotation: tuple
    length: float
    parentId: int

    def __init__(self, reader: EndianBinaryReader) -> None:
        self.name = reader.read_aligned_string()
        self.position = reader.read_vector3()
        self.rotation = reader.read_vector3()
        self.length = reader.read_float()
        self.parentId = reader.read_int()

    def save(self, writer: EndianBinaryWriter):
        writer.write_aligned_string(self.name)
        writer.write_vector3(self.position)
        writer.write_vector3(self.rotation)
        writer.write_float(self.length)
        writer.write_int(self.parentId)



from .Object import Object
from .PPtr import PPtr, save_ptr
from ..streams import EndianBinaryReader, EndianBinaryWriter


class ResourceManager(Object):
    def __init__(self, reader):
        super().__init__(reader)
        m_ContainerSize = reader.read_int()
        self.m_Container = {
            reader.read_aligned_string(): PPtr(reader)
            for _ in range(m_ContainerSize)
        }

    def save(self, writer: EndianBinaryWriter = None):
        if not writer:
            writer = EndianBinaryWriter(endian=self.reader.endian)
        super().save(writer, intern_call=True)
        writer.write_int(len(self.m_Container))
        for key, val in self.m_Container.items():
            writer.write_aligned_string(key)
            save_ptr(val, writer)
        return writer



from ..enums import FileType
from ..helpers import ImportHelper
from ..streams import EndianBinaryReader, EndianBinaryWriter

from collections import namedtuple
from os.path import basename

DirectoryInfo = namedtuple("DirectoryInfo", "path offset size")


class File(object):
    name: str
    files: dict
    cab_file: str
    is_changed: bool
    signature: str
    packer: str

    # parent: File
    # environment: Environment

    def __init__(self, parent=None, name=None):
        self.files = {}
        self.is_changed = False
        self.cab_file = "CAB-UnityPy_Mod.resS"
        self.parent = parent
        self.environment = self.environment = getattr(parent, "environment", parent) if parent else None
        self.name = basename(name) if isinstance(name, str) else ""

    def get_assets(self):
        if isinstance(self, SerializedFile.SerializedFile):
            return self

        for f in self.files.values():
            if isinstance(f, (BundleFile.BundleFile, WebFile.WebFile)):
                for asset in f.get_assets():
                    yield asset
            elif isinstance(f, SerializedFile.SerializedFile):
                yield f

    def get_filtered_objects(self, obj_types=[]):
        if len(obj_types) == 0:
            return self.get_objects()
        for f in self.files.values():
            if isinstance(f, (BundleFile.BundleFile, WebFile.WebFile)):
                for obj in f.objects:
                    if obj.type in obj_types:
                        yield obj
            elif isinstance(f, SerializedFile.SerializedFile):
                for obj in f.objects.values():
                    if obj.type in obj_types:
                        yield obj

    def get_objects(self):
        for f in self.files.values():
            if isinstance(f, (BundleFile.BundleFile, WebFile.WebFile)):
                for obj in f.objects:
                    yield obj
            elif isinstance(f, SerializedFile.SerializedFile):
                for obj in f.objects.values():
                    yield obj
            elif isinstance(f, ObjectReader.ObjectReader):
                yield f

    def read_files(self, reader: EndianBinaryReader, files: list):
        # read file data and convert it
        for node in files:
            reader.Position = node.offset
            name = node.path
            f = EndianBinaryReader(
                reader.read(node.size), offset=(reader.BaseOffset + node.offset)
            )
            # f._flag = getattr(node, "flags", None)  # required for save
            typ, _ = ImportHelper.check_file_type(f)
            if typ == FileType.BundleFile:
                f = BundleFile.BundleFile(f, self, name=name)
            elif typ == FileType.WebFile:
                f = WebFile.WebFile(f, self, name=name)
            elif typ == FileType.AssetsFile:
                # pre-check if resource file
                if not name.endswith((".resS", ".resource", ".config", ".xml", ".dat")):
                    # try to load the file as serialized file
                    try:
                        f = SerializedFile.SerializedFile(f, self, name=name)
                    except ValueError:
                        pass

            if isinstance(f, (EndianBinaryReader, SerializedFile.SerializedFile)):
                if self.environment:
                    self.environment.register_cab(name, f)

            # required for BundleFiles
            f.flags = getattr(node, "flags", 0)
            self.files[name] = f

    def get_writeable_cab(self, name: str = None):
        """
        Creates a new cab file in the bundle that contains the given data.
        This is usefull for asset types that use resource files.
        """

        if not name:
            name = self.cab_file

        if not name:
            return None

        if name in self.files:
            if isinstance(self.files[name], EndianBinaryWriter):
                return self.files[name]
            else:
                raise ValueError(
                    "This cab already exists and isn't an EndianBinaryWriter"
                )

        writer = EndianBinaryWriter()
        # try to find another resource file to copy the flags from
        for fname, f in self.files.items():
            if fname.endswith(".resS"):
                writer.flags = f.flags
                writer.endian = f.endian
                break
        else:
            writer.flags = 0
        writer.name = name
        self.files[name] = writer
        return writer

    @property
    def container(self):
        return {
            path: obj
            for f in self.files.values()
            if isinstance(f, File)
            for path, obj in f.container.items()
        }

    def get(self, key, default=None):
        return getattr(self, key, default)

    def keys(self):
        return self.files.keys()

    def items(self):
        return self.files.items()

    def values(self):
        return self.files.values()

    def __getitem__(self, item):
        return self.files[item]

    def __repr__(self):
        return f"<{self.__class__.__name__}>"

    def mark_changed(self):
        if isinstance(self.parent, File):
            self.parent.mark_changed()
        self.is_changed = True


# recursive import requires the import down here
from . import BundleFile, SerializedFile, WebFile, ObjectReader




from .Component import Component
from ..streams import EndianBinaryReader, EndianBinaryWriter

class Behaviour(Component):
    def __init__(self, reader : EndianBinaryReader):
        super().__init__(reader=reader)
        self.m_Enabled = reader.read_byte()
        reader.align_stream()
    
    def save(self, writer: EndianBinaryWriter = None):
        if writer is None:
            writer = EndianBinaryWriter(endian=self.reader.endian)
        version = self.version
        super().save(writer)
        writer.write_byte(self.m_Enabled)
        writer.align_stream()


import io
from struct import unpack

from ..math import Color, Matrix4x4, Quaternion, Vector2, Vector3, Vector4, Rectangle


class EndianBinaryReader:
    endian: str
    Length: int
    Position: int
    BaseOffset: int

    def __new__(cls, item, endian=">", offset=0):
        if isinstance(item, (bytes, bytearray, memoryview)):
            obj = super(EndianBinaryReader, cls).__new__(
                EndianBinaryReader_Memoryview)
        else:
            obj = super(EndianBinaryReader, cls).__new__(
                EndianBinaryReader_Streamable
            )
        obj.__init__(item, endian)
        return obj

    def __init__(self, item, endian=">", offset=0):
        self.endian = endian
        self.BaseOffset = offset
        self.Position = 0

    @property
    def bytes(self):
        # implemented by Streamable and Memoryview versions
        return b""

    def read(self, *args):
        # implemented by Streamable and Memoryview versions
        return b""

    def read_byte(self) -> int:
        return unpack(self.endian + "b", self.read(1))[0]

    def read_u_byte(self) -> int:
        return unpack(self.endian + "B", self.read(1))[0]

    def read_bytes(self, num) -> bytes:
        return self.read(num)

    def read_short(self) -> int:
        return unpack(self.endian + "h", self.read(2))[0]

    def read_int(self) -> int:
        return unpack(self.endian + "i", self.read(4))[0]

    def read_long(self) -> int:
        return unpack(self.endian + "q", self.read(8))[0]

    def read_u_short(self) -> int:
        return unpack(self.endian + "H", self.read(2))[0]

    def read_u_int(self) -> int:
        return unpack(self.endian + "I", self.read(4))[0]

    def read_u_long(self) -> int:
        return unpack(self.endian + "Q", self.read(8))[0]

    def read_float(self) -> float:
        return unpack(self.endian + "f", self.read(4))[0]

    def read_double(self) -> float:
        return unpack(self.endian + "d", self.read(8))[0]

    def read_boolean(self) -> bool:
        return bool(unpack(self.endian + "?", self.read(1))[0])

    def read_string(self, size=None, encoding="utf8") -> str:
        if size is None:
            ret = self.read_string_to_null()
        else:
            ret = unpack(f"{self.endian}{size}is", self.read(size))[0]
        try:
            return ret.decode(encoding)
        except UnicodeDecodeError:
            return ret

    def read_string_to_null(self, max_length=32767) -> str:
        ret = []
        c = b""
        while c != b"\0" and len(ret) < max_length and self.Position != self.Length:
            ret.append(c)
            c = self.read(1)
            if not c:
                raise ValueError("Unterminated string: %r" % ret)
        return b"".join(ret).decode("utf8", "surrogateescape")

    def read_aligned_string(self):
        length = self.read_int()
        if 0 < length <= self.Length - self.Position:
            string_data = bytes(self.read_bytes(length))
            result = string_data.decode("utf8", "surrogateescape")
            self.align_stream()
            return result
        return ""

    def align_stream(self, alignment=4):
        self.Position += (alignment - self.Position % alignment) % alignment

    def read_quaternion(self):
        return Quaternion(
            self.read_float(), self.read_float(), self.read_float(), self.read_float()
        )

    def read_vector2(self):
        return Vector2(self.read_float(), self.read_float())

    def read_vector3(self):
        return Vector3(self.read_float(), self.read_float(), self.read_float())

    def read_vector4(self):
        return Vector4(
            self.read_float(), self.read_float(), self.read_float(), self.read_float()
        )

    def read_rectangle_f(self):
        return Rectangle(
            self.read_float(), self.read_float(), self.read_float(), self.read_float()
        )

    def read_color4(self):
        return Color(
            self.read_float(), self.read_float(), self.read_float(), self.read_float()
        )

    def read_byte_array(self):
        return self.read(self.read_int())

    def read_matrix(self):
        return Matrix4x4(self.read_float_array(16))

    def read_array(self, command, length: int):
        return [command() for _ in range(length)]

    def read_boolean_array(self):
        return self.read_array(self.read_boolean, self.read_int())

    def read_u_short_array(self):
        return self.read_array(self.read_u_short, self.read_int())

    def read_int_array(self, length=0):
        return self.read_array(self.read_int, length if length else self.read_int())

    def read_u_int_array(self, length=0):
        return self.read_array(self.read_u_int, length if length else self.read_int())

    def read_u_int_array_array(self, length=0):
        return self.read_array(
            self.read_u_int_array, length if length else self.read_int()
        )

    def read_float_array(self, length=0):
        return self.read_array(self.read_float, length if length else self.read_int())

    def read_string_array(self):
        return self.read_array(self.read_aligned_string, self.read_int())

    def read_vector2_array(self):
        return self.read_array(self.read_vector2, self.read_int())

    def read_vector4_array(self):
        return self.read_array(self.read_vector4, self.read_int())

    def read_matrix_array(self):
        return self.read_array(self.read_matrix, self.read_int())

    def real_offset(self):
        """ Returns offset in the underlying file.
            (Not working with unpacked streams.)
        """
        return self.BaseOffset + self.Position

    def read_the_rest(self, obj_start, obj_size):
        """ Returns the rest of the current reader bytes."""
        return self.read_bytes(obj_size - (self.Position - obj_start))


class EndianBinaryReader_Memoryview(EndianBinaryReader):
    view: memoryview

    def __init__(self, view, endian=">", offset=0):
        super().__init__(view, endian=endian, offset=offset)
        self.view = memoryview(view)
        self.Length = len(view)

    @property
    def bytes(self):
        return self.view

    def dispose(self):
        self.view.release()

    def read(self, length: int):
        if not length:
            return b""
        ret = self.view[self.Position: self.Position + length]
        self.Position += length
        return ret

    def read_aligned_string(self):
        length = self.read_int()
        if 0 < length <= self.Length - self.Position:
            string_data = self.read_bytes(length)
            result = bytes(string_data).decode("utf8", "surrogateescape")
            self.align_stream()
            return result
        return ""


class EndianBinaryReader_Streamable(EndianBinaryReader):
    stream: io.BufferedReader

    def __init__(self, stream, endian=">", offset=0):
        self.stream = stream
        super().__init__(stream, endian=endian, offset=offset)

    def get_position(self):
        return self.stream.tell()

    def set_position(self, value):
        self.stream.seek(value+self.BaseOffset)

    @property
    def Length(self):
        pos = self.Position
        length = self.stream.seek(0, 2) - self.BaseOffset
        self.Position = pos
        return length

    Position = property(get_position, set_position)

    @property
    def bytes(self):
        last_pos = self.Position
        self.Position = 0
        ret = self.read(self.Length)
        self.Position = last_pos
        return ret

    def dispose(self):
        self.stream.close()
        pass

    def read(self, length: int):
        if not length:
            return b""
        return self.stream.read(length)




from .Behaviour import Behaviour
from .PPtr import PPtr, save_ptr
from ..streams import EndianBinaryReader, EndianBinaryWriter
from ..exceptions import TypeTreeError as TypeTreeError

class MonoBehaviour(Behaviour):
    def __init__(self, reader: EndianBinaryReader):
        super().__init__(reader=reader)
        self.m_Script = PPtr(reader)
        self.name = reader.read_aligned_string()

        self._raw_offset = reader.Position
        if self.assets_file._enable_type_tree:
            try:
                self.read_typetree()
            except TypeTreeError as e:
                print("Failed to read TypeTree:\n", e)
                self.assets_file._enable_type_tree = False

    def save(self, writer: EndianBinaryWriter = None, raw_data: bytes = None):
        if writer is None:
            writer = EndianBinaryWriter(endian=self.reader.endian)
        if not raw_data:
            ValueError("No raw data given")
        
        super().save(writer)
        save_ptr(self.m_Script, writer)
        writer.write_aligned_string(self.name)
        writer.write(raw_data)
        
        self.set_raw_data(writer.bytes)

    @property
    def raw_data(self) -> bytes:
        """
        Reads the undocumentated data following the default init.
        This is usefull for classes that are stored via MonoBehaviours.
        """
        reader = self.reader
        reader.Position = self._raw_offset
        return reader.read_bytes(reader.byte_size - (self._raw_offset - reader.byte_start))




from .EditorExtension import EditorExtension
from .PPtr import PPtr, save_ptr
from ..streams import EndianBinaryReader, EndianBinaryWriter


class Component(EditorExtension):
    def __init__(self, reader: EndianBinaryReader):
        super().__init__(reader=reader)
        self.m_GameObject = PPtr(reader)  # GameObject

    def save(self, writer: EndianBinaryWriter = None):
        if writer is None:
            writer = EndianBinaryWriter(endian=self.reader.endian)
        version = self.version
        super().save(writer)
        save_ptr(self.m_GameObject, writer)



class TypeTreeError(Exception):
    def __init__(self, message, nodes):            
        # Call the base class constructor with the parameters it needs
        super().__init__(message)
        self.nodes = nodes


from .NamedObject import NamedObject
from ..enums import AudioType, AudioCompressionFormat, AUDIO_TYPE_EXTEMSION
from ..export import AudioClipConverter
from ..helpers.ResourceReader import get_resource_data


class AudioClip(NamedObject):
    def __init__(self, reader):
        super().__init__(reader=reader)
        self.m_Source = ""
        version = self.version
        if version < (5,):  # 5.0 down
            self.m_Format = reader.read_int()
            self.m_Type = AudioType(reader.read_int())
            self.m_3D = reader.read_boolean()
            self.m_UseHardware = reader.read_boolean()
            reader.align_stream()

            if version >= (3, 2):  # and version <= (5,): # 3.2.0 to 5
                self.m_Stream = reader.read_int()
                self.m_Size = reader.read_int()
                tsize = self.m_Size + 4 - self.m_Size % 4 if (self.m_Size % 4 != 0) else self.m_Size
                if reader.byte_size + reader.byte_start - reader.Position != tsize:
                    self.m_Offset = reader.read_u_int()
                    self.m_Source = self.assets_file.full_name + ".resS"
            else:
                self.m_Size = reader.read_int()

        else:
            self.m_LoadType = reader.read_int()
            self.m_Channels = reader.read_int()
            self.m_Frequency = reader.read_int()
            self.m_BitsPerSample = reader.read_int()
            self.m_Length = reader.read_float()
            self.m_IsTrackerFormat = reader.read_boolean()
            reader.align_stream()
            self.m_SubsoundIndex = reader.read_int()
            self.m_PreloadAudioData = reader.read_boolean()
            self.m_LoadInBackground = reader.read_boolean()
            self.m_Legacy3D = reader.read_boolean()
            reader.align_stream()
            self.m_Source = reader.read_aligned_string()
            self.m_Offset = reader.read_u_long()
            self.m_Size = reader.read_long()
            self.m_CompressionFormat = AudioCompressionFormat(reader.read_int())

        if self.m_Source:
            self.m_AudioData = get_resource_data(
                self.m_Source, self.assets_file, self.m_Offset, self.m_Size
            )
        else:
            self.m_AudioData = reader.read_bytes(self.m_Size)
    
    @property
    def extension(self):
        return  AUDIO_TYPE_EXTEMSION.get(self.m_CompressionFormat, ".audioclip")

    @property
    def samples(self) -> dict:
        return AudioClipConverter.extract_audioclip_samples(self)



import setuptools
from UnityPy import __version__ as version

with open("README.md", "r") as fh:
	long_description = fh.read()

setuptools.setup(
	name="UnityPy",
	packages=setuptools.find_packages(),
	include_package_data = True,
	version=version,
	author="K0lb3",
	description="A pythonic port of AssetStudio by Perfare",
	long_description=long_description,
	long_description_content_type="text/markdown",
	url="https://github.com/K0lb3/UnityPy",
	download_url="https://github.com/K0lb3/UnityPy/tarball/master",
	keywords=['python', 'unity', 'unity-asset', 'python3', 'data-minig', 'unitypack', 'assetstudio', 'unity-asset-extractor'],
	classifiers=[
		"License :: OSI Approved :: MIT License",
		"Operating System :: OS Independent",
		"Intended Audience :: Developers",
		"Development Status :: 5 - Production/Stable",
		"Programming Language :: Python",
		"Programming Language :: Python :: 3",
		"Programming Language :: Python :: 3.6",
		"Programming Language :: Python :: 3.7",
		"Programming Language :: Python :: 3.8",
		"Programming Language :: Python :: 3.9",
		"Topic :: Software Development :: Libraries :: Python Modules",
		"Topic :: Games/Entertainment",
		"Topic :: Multimedia :: Graphics",
	],
	install_requires=[
		# block compression/decompression
		"lz4", # BundleFile block compression
		"brotli", # WebFile compression
		# Texture & Sprite handling
		"Pillow",
		"texture2ddecoder", # texture decompression
		"etcpak", # ETC & DXT compression
		# raw typetree dumping
		"tabulate"
	]
)



from typing import Any, Dict, List, Union
from UnityPy.streams import EndianBinaryReader, EndianBinaryWriter
from ctypes import c_uint32
import tabulate
from ..exceptions import TypeTreeError as TypeTreeError


class TypeTreeNode(object):
    type: str
    name: str
    byte_size: int
    index: int
    is_array: int
    version: int
    meta_flag: int
    level: int
    type_str_offset: int
    name_str_offset: int

    def __init__(self, data: dict = None):
        if data:
            self.__dict__.update(data)

    def __repr__(self):
        return f"<TypeTreeNode({self.level} {self.type} {self.name})>"


def node_dict_to_class(nodes: List[dict]) -> List[TypeTreeNode]:
    """Converts all dict-type nodes into TypeTreeNodes

        Parameters
        ----------
        nodes : List[dict]
            nodes/nodes of the typetree as dict

        Returns
        -------
        List[TypeTreeNode]
            a list of TypeTreeNode-type nodes
    """
    return [TypeTreeNode(node) for node in nodes]


def check_nodes(nodes: List[Union[dict, TypeTreeNode]]) -> List[TypeTreeNode]:
    """Checks the type of the nodes and converts them if necessary.

        Parameters
        ----------
        nodes : List[Union[dict, TypeTreeNode]]
            nodes/nodes of the typetree as dict or TypeTreeNode

        Returns
        -------
        List[TypeTreeNode]
            a list of TypeTreeNode-type nodes
    """
    if isinstance(nodes, list):
        if len(nodes) == 0:
            raise ValueError("not enough nodes")
        if isinstance(nodes[0], TypeTreeNode):
            return nodes
        elif isinstance(nodes[0], dict):
            return node_dict_to_class(nodes)
    raise ValueError(
        f"nodes must be a list of dict or TypeTreeNode elements, but received {type(nodes)} - {type(nodes[0]) if isinstance(nodes, list) else ''}"
    )


"""
Example Nodes:
[{
    "level": 0,
    "type": "MonoBehaviour",
    "name": "Base",
    "meta_flag": 0
},
{
    "level": 1,
    "type": "int",
    "name": "m_SomeNode",
    "meta_flag": 0
}]

Example TypeTree:
{
    "m_SomeNode": 1
}
"""


def get_nodes(nodes: List[TypeTreeNode], index: int) -> list:
    """Copies all nodes above the level of the node at the set index.

        Parameters
        ----------
        nodes : list
            nodes/nodes of the typetree
        index : int
            index of the node

        Returns
        -------
        list
            A list of nodes
    """
    nodes2 = [nodes[index]]
    level = nodes[index].level
    for node in nodes[index + 1 :]:
        if node.level <= level:
            return nodes2
        nodes2.append(node)
    return nodes2


def read_typetree(
    nodes: List[Union[dict, TypeTreeNode]], reader: EndianBinaryReader
) -> dict:
    """Reads the typetree of the object contained in the reader via the node list.

        Parameters
        ----------
        nodes : list
            List of nodes/nodes
        reader : EndianBinaryReader
            Reader of the object to be parsed

        Returns
        -------
        dict
            The parsed typtree
    """
    # reader.reset()
    nodes = check_nodes(nodes)

    obj = {}
    i = c_uint32(1)
    while i.value < len(nodes):
        node = nodes[i.value]
        obj[node.name] = read_value(nodes, reader, i)
        i.value += 1

    readed = reader.Position - reader.byte_start
    if readed != reader.byte_size:
        raise TypeTreeError(
            f"Error while read type, read {readed} bytes but expected {reader.byte_size} bytes",
            nodes,
        )

    return obj


def read_value(nodes: List[TypeTreeNode], reader: EndianBinaryReader, i: c_uint32):
    node = nodes[i.value]
    typ = node.type
    align = (node.meta_flag & 0x4000) != 0

    if typ == "SInt8":
        value = reader.read_byte()
    elif typ in ["UInt8", "char"]:
        value = reader.read_u_byte()
    elif typ in ["short", "SInt16"]:
        value = reader.read_short()
    elif typ in ["UInt16", "unsigned short"]:
        value = reader.read_u_short()
    elif typ in ["int", "SInt32"]:
        value = reader.read_int()
    elif typ in ["UInt32", "unsigned int", "Type*"]:
        value = reader.read_u_int()
    elif typ in ["long long", "SInt64"]:
        value = reader.read_long()
    elif typ in ["UInt64", "unsigned long long", "FileSize"]:
        value = reader.read_u_long()
    elif typ == "float":
        value = reader.read_float()
    elif typ == "double":
        value = reader.read_double()
    elif typ == "bool":
        value = reader.read_boolean()
    elif typ == "string":
        value = reader.read_aligned_string()
        i.value += 3  # Array, Size, Data(typ)
    elif typ == "map":  # map == MultiDict
        if (nodes[i.value + 1].meta_flag & 0x4000) != 0:
            align = True
        map_ = get_nodes(nodes, i.value)
        i.value += len(map_) - 1
        first = get_nodes(map_, 4)
        second = get_nodes(map_, 4 + len(first))
        size = reader.read_int()
        value = [None] * size
        for j in range(size):
            key = read_value(first, reader, c_uint32(0))
            value[j] = (key, read_value(second, reader, c_uint32(0)))
    elif typ == "TypelessData":
        size = reader.read_int()
        value = reader.read_bytes(size)
        i.value += 2  # Size == int, Data(typ) == char/uint8
    else:
        # Vector
        if i.value < len(nodes) - 1 and nodes[i.value + 1].type == "Array":
            if (nodes[i.value + 1].meta_flag & 0x4000) != 0:
                align = True
            vector = get_nodes(nodes, i.value)
            i.value += len(vector) - 1
            size = reader.read_int()
            value = [read_value(vector, reader, c_uint32(3)) for _ in range(size)]
        else:  # Class
            clz = get_nodes(nodes, i.value)
            i.value += len(clz) - 1
            value = {}
            j = c_uint32(1)
            while j.value < len(clz):
                clz_node = clz[j.value]
                value[clz_node.name] = read_value(clz, reader, j)
                j.value += 1

    if align:
        reader.align_stream()
    return value


def read_typetree_str(
    sb: List[str], nodes: List[Union[dict, TypeTreeNode]], reader: EndianBinaryReader
) -> list:
    """Reads the typetree of the object contained in the reader via the node list and dumps it as string.

        Parameters
        ----------
        sb : list
            StringBuilder - a list used to build the string dump, should be empty
        nodes : list
            List of nodes/nodes
        reader : EndianBinaryReader
            Reader of the object to be parsed

        Returns
        -------
        list
            The sb given as input
    """
    # reader.reset()
    nodes = check_nodes(nodes)

    i = c_uint32(0)
    while i.value < len(nodes):
        read_value_str(sb, nodes, reader, i)
        i.value += 1

    readed = reader.Position - reader.byte_start
    if readed != reader.byte_size:
        raise TypeTreeError(
            f"Error while read type, read {readed} bytes but expected {reader.byte_size} bytes",
            nodes,
        )

    return sb


def read_value_str(
    sb: List[str], nodes: List[TypeTreeNode], reader: EndianBinaryReader, i: c_uint32
) -> list:
    node = nodes[i.value]
    typ = node.type
    align = (node.meta_flag & 0x4000) != 0
    append = True

    if typ == "SInt8":
        value = reader.read_byte()
    elif typ in ["UInt8", "char"]:
        value = reader.read_u_byte()
    elif typ in ["short", "SInt16"]:
        value = reader.read_short()
    elif typ in ["UInt16", "unsigned short"]:
        value = reader.read_u_short()
    elif typ in ["int", "SInt32"]:
        value = reader.read_int()
    elif typ in ["UInt32", "unsigned int", "Type*"]:
        value = reader.read_u_int()
    elif typ in ["long long", "SInt64"]:
        value = reader.read_long()
    elif typ in ["UInt64", "unsigned long long", "FileSize"]:
        value = reader.read_u_long()
    elif typ == "float":
        value = reader.read_float()
    elif typ == "double":
        value = reader.read_double()
    elif typ == "bool":
        value = reader.read_boolean()
    elif typ == "string":
        value = reader.read_aligned_string()
        i.value += 3  # Array, Size, Data(typ)
        append = False
        sb.append(
            '{0}{1} {2} = "{3}"\r\n'.format(
                "\t" * node.level, node.type, node.name, value
            )
        )
    elif typ == "map":
        if (nodes[i.value + 1].meta_flag & 0x4000) != 0:
            align = True
        map_ = get_nodes(nodes, i.value)
        i.value += len(map_) - 1
        first = get_nodes(map_, 4)
        second = get_nodes(map_, 4 + len(first))
        size = reader.read_int()
        append = False
        sb.append("{0}{1} {2}\r\n".format("\t" * node.level, node.type, node.name))
        sb.append("{0}{1} {2}\r\n".format("\t" * (node.level + 1), "Array", "Array"))
        sb.append(
            "{0}{1} {2} = {3}\r\n".format("\t" * (node.level + 1), "int", "size", size)
        )
        for j in range(size):
            sb.append("{0}[{1}]\r\n".format("\t" * (node.level + 2), j))
            sb.append("{0}{1} {2}\r\n".format("\t" * (node.level + 2), "pair", "data"))
            read_value_str(sb, first, reader, c_uint32(0))
            read_value_str(sb, second, reader, c_uint32(0))
    elif typ == "TypelessData":
        size = reader.read_int()
        value = reader.read_bytes(size)
        i.value += 2  # Size == int, Data(typ) == char/uint8
        append = False
        sb.append("{0}{1} {2}\r\n".format("\t" * node.level, node.type, node.name))
        sb.append("{0}{1} {2} = {3}\r\n".format("\t" * node.level, "int", "size", size))
        # sb.append("{0}{1} {2} = {3}\r\n".format(
        #    "\t" * node.level, "UInt8", "data", base64.b64encode(value)))
    else:
        # Vector
        if i.value < len(nodes) - 1 and nodes[i.value + 1].type == "Array":
            if (nodes[i.value + 1].meta_flag & 0x4000) != 0:
                align = True
            vector = get_nodes(nodes, i.value)
            i.value += len(vector) - 1
            size = reader.read_int()
            append = False
            sb.append("{0}{1} {2}\r\n".format("\t" * node.level, node.type, node.name))
            sb.append(
                "{0}{1} {2}\r\n".format("\t" * (node.level + 1), "Array", "Array")
            )
            sb.append(
                "{0}{1} {2} = {3}\r\n".format(
                    "\t" * (node.level + 1), "int", "size", size
                )
            )
            for j in range(size):
                sb.append("{0}[{1}]\r\n".format("\t" * (node.level + 2), j))
                read_value_str(sb, vector, reader, c_uint32(3))

        else:  # Class
            clz = get_nodes(nodes, i.value)
            i.value += len(clz) - 1
            j = c_uint32(1)
            append = False
            sb.append("{0}{1} {2}\r\n".format("\t" * node.level, node.type, node.name))
            while j.value < len(clz):
                read_value_str(sb, clz, reader, j)
                j.value += 1

    if append:
        sb.append(
            "{0}{1} {2} = {3}\r\n".format(
                "\t" * node.level, node.type, node.name, value
            )
        )

    if align:
        reader.align_stream()
    return sb


def dump_typetree(nodes: List[TypeTreeNode]) -> str:
    """Dumps the structure of the given nodes.

        Parameters
        ----------
        nodes : list
            List of nodes/nodes

        Returns
        -------
        str
            The dumped structure
    """
    field_names = ["level", "type", "name", "meta_flag"]
    rows = [[getattr(x, key) for key in field_names] for x in nodes]
    return tabulate.tabulate(rows, headers=field_names)


def write_typetree(
    obj: dict, nodes: List[Union[dict, TypeTreeNode]], writer: EndianBinaryWriter = None
) -> EndianBinaryWriter:
    """Writes the data of the object via the given typetree of the object into the writer.

        Parameters
        ----------
        obj : dict
            Object to be saved
        nodes : list
            List of nodes/nodes
        writer : EndianBinaryWriter
            Writer of the object to be saved

        Returns
        -------
        EndianBinaryWriter
            The writer that was used to save the data of the given object.
    """
    if not writer:
        writer = EndianBinaryWriter()

    nodes = check_nodes(nodes)

    i = c_uint32(1)
    while i.value < len(nodes):
        value = obj[nodes[i.value].name]
        write_value(value, nodes, writer, i)
        i.value += 1
    return writer


def write_value(
    value: Any, nodes: List[TypeTreeNode], writer: EndianBinaryWriter, i: c_uint32
):
    node = nodes[i.value]
    typ = node.type
    align = (node.meta_flag & 0x4000) != 0

    if typ == "SInt8":
        writer.write_byte(value)
    elif typ in ["UInt8", "char"]:
        writer.write_u_byte(value)
    elif typ in ["short", "SInt16"]:
        writer.write_short(value)
    elif typ in ["UInt16", "unsigned short"]:
        writer.write_u_short(value)
    elif typ in ["int", "SInt32"]:
        writer.write_int(value)
    elif typ in ["UInt32", "unsigned int", "Type*"]:
        writer.write_u_int(value)
    elif typ in ["long long", "SInt64"]:
        writer.write_long(value)
    elif typ in ["UInt64", "unsigned long long", "FileSize"]:
        writer.write_u_long(value)
    elif typ == "float":
        writer.write_float(value)
    elif typ == "double":
        writer.write_double(value)
    elif typ == "bool":
        writer.write_boolean(value)
    elif typ == "string":
        writer.write_aligned_string(value)
        i.value += 3  # Array, Size, Data(typ)
    elif typ == "map":
        if (nodes[i.value + 1].meta_flag & 0x4000) != 0:
            align = True
        map_ = get_nodes(nodes, i.value)
        i.value += len(map_) - 1
        first = get_nodes(map_, 4)
        second = get_nodes(map_, 4 + len(first))
        # size
        writer.write_int(len(value))
        # data
        for key, val in value:
            write_value(key, first, writer, c_uint32(0))
            write_value(val, second, writer, c_uint32(0))

    elif typ == "TypelessData":
        writer.write_int(len(value))
        writer.write_bytes(value)
        i.value += 2  # Size == int, Data(typ) == char/uint8
    else:
        # Vector
        if i.value < len(nodes) - 1 and nodes[i.value + 1].type == "Array":
            if (nodes[i.value + 1].meta_flag & 0x4000) != 0:
                align = True
            vector = get_nodes(nodes, i.value)
            i.value += len(vector) - 1
            writer.write_int(len(value))
            for val in value:
                write_value(val, vector, writer, c_uint32(3))
        else:  # Class
            clz = get_nodes(nodes, i.value)
            i.value += len(clz) - 1
            j = c_uint32(1)
            while j.value < len(clz):
                val = value[clz[j.value].name]
                write_value(val, clz, writer, j)
                j.value += 1

    if align:
        writer.align_stream()




from tkinter.messagebox import NO
from .EditorExtension import EditorExtension
from .PPtr import PPtr
from ..enums import ClassIDType


class GameObject(EditorExtension):
    m_Components: list
    m_Layer: int
    name: str
    m_Animator: PPtr
    m_Animation: PPtr
    m_Transform: PPtr
    m_MeshRenderer: PPtr
    m_SkinnedMeshRender: PPtr
    m_MeshFilter: PPtr

    def __init__(self, reader):
        super().__init__(reader=reader)

        self.m_Animator = None
        self.m_Animation = None
        self.m_Transform = None
        self.m_MeshRenderer = None
        self.m_SkinnedMeshRenderer = None
        self.m_MeshFilter = None

        component_size = reader.read_int()

        self.m_Components = [None] * component_size
        for i in range(component_size):
            if self.version < (5, 5):
                first = reader.read_int()
            component = PPtr(reader)
            self.m_Components[i] = component

            if component.type == ClassIDType.Animator:
                self.m_Animator = component
            elif component.type == ClassIDType.Animation:
                self.m_Animation = component
            elif component.type == ClassIDType.Transform:
                self.m_Transform = component
            elif component.type == ClassIDType.MeshRenderer:
                self.m_MeshRenderer = component
            elif component.type == ClassIDType.SkinnedMeshRenderer:
                self.m_SkinnedMeshRenderer = component
            elif component.type == ClassIDType.MeshFilter:
                self.m_MeshFilter = component

        self.m_Layer = reader.read_int()
        self.name = reader.read_aligned_string()



from .NamedObject import NamedObject
from ..helpers.ResourceReader import get_resource_data
from .PPtr import PPtr


class VideoClip(NamedObject):
    def __init__(self, reader):
        super().__init__(reader=reader)
        self.m_OriginalPath = reader.read_aligned_string()
        self.m_ProxyWidth = reader.read_u_int()
        self.m_ProxyHeight = reader.read_u_int()
        self.Width = reader.read_u_int()
        self.Height = reader.read_u_int()
        if self.version >= (2017, 2):  # 2017.2 and up
            self.m_PixelAspecRatioNum = reader.read_u_int()
            self.m_PixelAspecRatioDen = reader.read_u_int()
        self.m_FrameRate = reader.read_double()
        self.m_FrameCount = reader.read_u_long()
        self.m_Format = reader.read_int()
        self.m_AudioChannelCount = reader.read_u_short_array()
        reader.align_stream()
        self.m_AudioSampleRate = reader.read_u_int_array()
        self.m_AudioLanguage = reader.read_string_array()
        if self.version >= (2020,):  # 2020.1 and up
            m_VideoShadersSize = reader.read_int()
            self.m_VideoShaders = [
                PPtr(reader) for _ in range(m_VideoShadersSize)
            ]

        # m_ExternalResources = new StreamedResource(reader)
        self.source = reader.read_aligned_string()
        self.offset = reader.read_u_long()
        self.size = reader.read_u_long()

        self.m_HasSplitAlpha = reader.read_boolean()
        if self.version >= (2020,):  # 2020.1 and up
            self.m_sRGB = reader.read_boolean()

        if self.source:
            self.m_VideoData = get_resource_data(
                self.source, self.assets_file, self.offset, self.size)
        else:
            self.reader.Position = self.data_offset
            self.m_VideoData = self.reader.read_bytes(self.size)



from functools import reduce
import io
import os
from zipfile import ZipFile
import re
from . import files
from .files import File
from .enums import FileType
from .helpers import ImportHelper
from .streams import EndianBinaryReader
from .files import SerializedFile

reSplitFile = re.compile(r"^(.+)\.split(\d+)$")


class Environment:
    files: dict
    cabs: dict
    path: str

    def __init__(self, *args):
        self.files = {}
        self.cabs = {}
        self.path = None
        self.out_path = os.path.join(os.getcwd(), "output")

        if args:
            for arg in args:
                if isinstance(arg, str):
                    if os.path.isfile(arg):
                        if os.path.splitext(arg)[-1] in [".apk", ".zip"]:
                            self.load_zip_file(arg)
                        else:
                            self.path = os.path.dirname(arg)
                            self.load_files([arg])
                    elif os.path.isdir(arg):
                        self.path = arg
                        self.load_folder(arg)
                else:
                    self.path = None
                    self.files[str(len(self.files))] = self.load_file(stream=arg)

        if len(self.files) == 1:
            self.file = list(self.files.values())[0]
        
        if self.path == "":
            self.path = os.getcwd()

    def load_files(self, files: list):
        """Loads all files (list) into the AssetsManager and merges .split files for common usage."""
        # filter split files
        i = 0
        splits = []
        while i < len(files):
            path = files[i]
            splitMatch = reSplitFile.match(path)
            if reSplitFile.match(path):
                name = splitMatch[1]
                if name not in splits:
                    splits.append(name)
                files.remove(path)
            else:
                i += 1

        # merge splits
        for basepath in splits:
            # merge data
            data = io.BytesIO()
            for i in range(0, 999):
                item = f"{basepath}.split{i}"
                if os.path.exists(item):
                    with open(item, "rb") as f:
                        data.write(f.read())
                elif i:
                    break
            self.files[self.reduce_path(basepath)] = self.load_file(data)

        # load all other files
        self.load(files)

    def load_folder(self, path: str):
        """Loads all files in the given path and its subdirs into the AssetsManager."""
        self.load_files(
            [
                os.path.join(root, f)
                for root, dirs, files in os.walk(path)
                for f in files
            ]
        )

    def reduce_path(self, fp: str) -> str:
        return fp[len(self.path) :].lstrip("/\\")

    def load(self, files: list):
        """Loads all files into the AssetsManager."""
        self.files.update(
            {
                self.reduce_path(f): self.load_file(open(f, "rb"))
                for f in files
                if os.path.exists(f)
            }
        )

    def load_file(self, stream, parent=None):
        if not parent:
            parent = self
        typ, reader = ImportHelper.check_file_type(stream)
        try:
            stream_name = getattr(
                stream,
                "name",
                str(stream.__hash__()) if hasattr(stream, "__hash__") else "",
            )
            if typ == FileType.AssetsFile:
                return files.SerializedFile(reader, parent, name=stream_name)
            elif typ == FileType.BundleFile:
                return files.BundleFile(reader, parent, name=stream_name)
            elif typ == FileType.WebFile:
                return files.WebFile(reader, parent, name=stream_name)
            elif typ == FileType.ZIP:
                self.load_zip_file(stream)
            elif typ == FileType.ResourceFile:
                return EndianBinaryReader(stream)
        except Exception as e:
            # just to be sure
            # cuz the SerializedFile detection isn't perfect
            print("Error loading, reverting to EndianBinaryReader:\n", e)
            return EndianBinaryReader(stream)

    def load_zip_file(self, value):
        buffer = None
        if isinstance(value, str) and os.path.exists(value):
            buffer = open(value, "rb")
        elif isinstance(value, (bytes, bytearray)):
            buffer = io.BytesIO(value)
        elif isinstance(value, (io.BufferedReader, io.BufferedIOBase)):
            buffer = value

        z = ZipFile(buffer)
        splits = []
        for path in z.namelist():
            splitMatch = reSplitFile.match(path)
            if reSplitFile.match(path):
                name = splitMatch[1]
                if name not in splits:
                    splits[name] = []
                continue

            stream = z.open(path)
            if stream._orig_file_size == 0:
                continue
            *path, name = path.split("/")
            cur = self
            for d in path:
                if d not in cur.files:
                    cur.files[d] = files.File(self)
                cur = cur.files[d]
            cur.files[name] = self.load_file(stream, cur)

        # merge splits
        for basepath in splits:
            # merge data
            data = io.BytesIO()
            for i in range(0, 999):
                item = f"{basepath}.split{i}"
                if item in z.namelist():
                    with z.open(item) as f:
                        data.write(f.read())
                elif i:
                    break
            *path, name = basepath.split("/")
            cur = self
            for d in path:
                if d not in cur.files:
                    cur.files[d] = files.File(self)
                cur = cur.files[d]
            cur.files[name] = self.load_file(data, cur)

        z.close()

    def save(self, pack="none"):
        """Saves all changed assets.
        Mark assets as changed using `.mark_changed()`.
        pack = "none" (default) or "lz4"
        """
        for f in self.files:
            if self.files[f].is_changed:
                with open(
                    os.path.join(self.out_path, os.path.basename(f)), "wb"
                ) as out:
                    out.write(self.files[f].save(packer=pack))

    @property
    def objects(self) -> list:
        def search(item):
            ret = []
            if not isinstance(item, Environment) and getattr(item, "objects", None):
                # serialized file
                return [val for val in item.objects.values()]

            elif getattr(item, "files", None):  # WebBundle and BundleFile
                # bundle
                for item in item.files.values():
                    ret.extend(search(item))
                return ret

            return ret

        return search(self)

    @property
    def container(self):
        return {
            path: obj
            for f in self.files.values()
            if isinstance(f, File)
            for path, obj in f.container.items()
        }

    @property
    def assets(self) -> list:
        """
        Lists all assets / SerializedFiles within this environment.
        """

        def gen_all_asset_files(file, ret=[]):
            for f in getattr(file, "files", {}).values():
                if isinstance(f, SerializedFile):
                    ret.append(f)
                else:
                    gen_all_asset_files(f, ret)
            return ret

        return gen_all_asset_files(self)

    def get(self, key: str, default=None):
        return getattr(self, key, default)

    def register_cab(self, name: str, item: object) -> None:
        self.cabs[name.lower()] = item

    def get_cab(self, name: str) -> object:
        return self.cabs.get(name.lower(), None)



import os
from ..classes import Renderer, SkinnedMeshRenderer, Material, Texture2D
from ..enums import ClassIDType
from .MeshExporter import export_mesh_obj


def get_mesh(meshR: Renderer):
    if isinstance(meshR, SkinnedMeshRenderer.SkinnedMeshRenderer):
        if meshR.m_Mesh:
            return meshR.m_Mesh.read()
    else:
        m_GameObject = meshR.m_GameObject.read()
        if m_GameObject.m_MeshFilter:
            return m_GameObject.m_MeshFilter.read().m_Mesh.read()
    return None


def export_mesh_renderer(obj: Renderer, export_dir: str) -> None:
    os.makedirs(export_dir, exist_ok=True)
    meshR = obj.read()
    mesh = get_mesh(meshR)
    if not mesh:
        return

    firstSubMesh = 0
    if hasattr(meshR, "m_StaticBatchInfo") and meshR.m_StaticBatchInfo.subMeshCount > 0:
        firstSubMesh = meshR.m_StaticBatchInfo.firstSubMesh
    elif hasattr(meshR, "m_SubsetIndices"):
        firstSubMesh = min(meshR.m_SubsetIndices)

    materials = []
    material_names = []
    for i, submesh in enumerate(mesh.m_SubMeshes):
        mat = None
        if i - firstSubMesh < len(meshR.m_Materials):
            matPtr = meshR.m_Materials[i - firstSubMesh]
            if matPtr:
                mat = matPtr.read()
        if not mat:
            material_names.append(None)
            continue
        materials.append(export_material(mat))
        material_names.append(mat.m_Name)
        # save material textures
        for key, texEnv in mat.m_SavedProperties.m_TexEnvs.items():
            if not texEnv.m_Texture:
                continue
            tex = texEnv.m_Texture.read()
            texName = f"{tex.m_Name if tex.m_Name else key}.png"
            tex.read().image.save(os.path.join(export_dir, texName))

    # save .obj
    with open(
        os.path.join(export_dir, f"{mesh.name}.obj"), "wt", encoding="utf8", newline=""
    ) as f:
        f.write(export_mesh_obj(mesh, material_names))

    # save .mtl
    with open(
        os.path.join(export_dir, f"{mesh.name}.mtl"), "wt", encoding="utf8", newline=""
    ) as f:
        f.write("\n".join(materials))


def export_material(mat: Material) -> str:
    """Creates a material file (.mtl) for the given material."""

    def clt(color):  # color to tuple
        return (
            color if isinstance(color, tuple) else (color.R, color.G, color.B, color.A)
        )

    colors = mat.m_SavedProperties.m_Colors
    floats = mat.m_SavedProperties.m_Floats
    texEnvs = mat.m_SavedProperties.m_TexEnvs

    diffuse = clt(colors.get("_Color", (0.8, 0.8, 0.8, 1)))
    ambient = clt(colors.get("_SColor", (0.2, 0.2, 0.2, 1)))
    emissive = clt(colors.get("_EmissionColor", (0, 0, 0, 1)))
    specular = clt(colors.get("_SpecularColor", (0.2, 0.2, 0.2, 1)))
    reflection = clt(colors.get("_ReflectColor", (0, 0, 0, 1)))
    shininess = floats.get("_Shininess", 20.0)
    transparency = floats.get("_Transparency", 0.0)

    sb = []
    sb.append(f"newmtl {mat.m_Name}")
    # Ka r g b
    # defines the ambient color of the material to be (r,g,b). The default is (0.2,0.2,0.2);
    sb.append(f"Ka {ambient[0]:.4f} {ambient[1]:.4f} {ambient[2]:.4f}")
    # Kd r g b
    # defines the diffuse color of the material to be (r,g,b). The default is (0.8,0.8,0.8);
    sb.append(f"Kd {diffuse[0]:.4f} {diffuse[1]:.4f} {diffuse[2]:.4f}")
    # Ks r g b
    # defines the specular color of the material to be (r,g,b). This color shows up in highlights. The default is (1.0,1.0,1.0);
    sb.append(f"Ks {specular[0]:.4f} {specular[1]:.4f} {specular[2]:.4f}")
    # d alpha
    # defines the non-transparency of the material to be alpha. The default is 1.0 (not transparent at all). The quantities d and Tr are the opposites of each other, and specifying transparency or nontransparency is simply a matter of user convenience.
    # Tr alpha
    # defines the transparency of the material to be alpha. The default is 0.0 (not transparent at all). The quantities d and Tr are the opposites of each other, and specifying transparency or nontransparency is simply a matter of user convenience.
    sb.append(f"Tr {transparency:.4f}")
    # Ns s
    # defines the shininess of the material to be s. The default is 0.0;
    sb.append(f"Ns {shininess:.4f}")
    # illum n
    # denotes the illumination model used by the material. illum = 1 indicates a flat material with no specular highlights, so the value of Ks is not used. illum = 2 denotes the presence of specular highlights, and so a specification for Ks is required.
    # map_Ka filename
    # names a file containing a texture map, which should just be an ASCII dump of RGB values;
    texName = None
    tex = None
    for key, texEnv in texEnvs.items():
        if not texEnv.m_Texture:
            continue
        tex = texEnv.m_Texture.read()
        texName = f"{tex.m_Name if tex.m_Name else key}.png"
        if key == "_MainTex":
            sb.append(f"map_Kd {texName}")
        elif key == "_BumpMap":
            # TODO: bump is default, some use map_bump
            sb.append(f"map_bump {texName}")
            sb.append(f"bump {texName}")
        elif "Specular" in key:
            sb.append(f"map_Ks {texName}")
        elif "Normal" in key:
            # TODO: figure out the key
            pass
    ret = "\n".join(sb)
    return ret



from ..files import ObjectReader
from ..streams import EndianBinaryWriter
from ..helpers import ImportHelper
from .. import files
from ..enums import FileType, ClassIDType
import os
from .. import environment


def save_ptr(obj, writer: EndianBinaryWriter):
    if isinstance(obj, PPtr):
        writer.write_int(obj.file_id)
    else:
        writer.write_int(0)  # it's usually 0......
    if obj._version < 14:
        writer.write_int(obj.path_id)
    else:
        writer.write_long(obj.path_id)


class PPtr:
    def __init__(self, reader: ObjectReader):
        self._version = reader.version2
        self.index = -2
        self.file_id = reader.read_int()
        self.path_id = reader.read_int() if self._version < 14 else reader.read_long()
        self.assets_file = reader.assets_file
        self._obj = None

    def save(self, writer: EndianBinaryWriter):
        save_ptr(self, writer)

    def get_obj(self):
        if self._obj != None:
            return self._obj
        manager = None
        if self.file_id == 0:
            manager = self.assets_file

        elif self.file_id > 0 and self.file_id - 1 < len(self.assets_file.externals):
            if self.index == -2:
                environment = self.assets_file.environment
                external_name = self.external_name
                # try to find it in the already registered cabs
                manager = environment.get_cab(external_name)

                if not manager:
                    # guess we have to try to find it as file then
                    path = environment.path
                    if path is not None:
                        basename = os.path.basename(external_name)
                        possible_names = [basename, basename.lower(), basename.upper()]
                        for root, dirs, files in os.walk(path):
                            for name in files:
                                if name in possible_names:
                                    manager = environment.load_file(
                                        os.path.join(root, name)
                                    )
                                    environment.register_cab(name, manager)
                                    break
                            else:
                                # else is reached if the previous loop didn't break
                                continue
                            break
        if manager and self.path_id in manager.objects:
            self._obj = manager.objects[self.path_id]
        else:
            self._obj = None
            if self.external_name:
                print(f"Couldn't find dependency {self.external_name}")
                print("You can try to load it manually to the environment in advance")
                print("for Web-&BundleFiles: env.load_file(dependency)")
                print(
                    "for SerializedFiles: env.register_cab(depdency_basename, env.load_file(dependency)"
                )
            else:
                print(f"Couldn't find referenced object with path_id {self.path_id}")

        return self._obj

    @property
    def type(self):
        obj = self.get_obj()
        if obj is None:
            return ClassIDType.UnknownType
        return obj.type

    @property
    def external_name(self):
        if self.file_id > 0 and self.file_id - 1 < len(self.assets_file.externals):
            return self.assets_file.externals[self.file_id - 1].name

    def __getattr__(self, key):
        obj = self.get_obj()
        return getattr(obj, key)

    def __repr__(self):
        return "<%s %s>" % (
            self.__class__.__name__,
            self._obj.__class__.__repr__(self.get_obj())
            if self.get_obj()
            else "Not Found",
        )

    def __bool__(self):
        return True if self.get_obj() else False



from enum import IntEnum


class GfxPrimitiveType(IntEnum):
    kPrimitiveTriangles = 0
    kPrimitiveTriangleStrip = 1
    kPrimitiveQuads = 2
    kPrimitiveLines = 3
    kPrimitiveLineStrip = 4
    kPrimitivePoints = 5